#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path

import onnx
import tensorrt as trt
from onnx import TensorProto, helper


def trt_dtype_to_onnx(dtype):
    if dtype == trt.float16:
        return TensorProto.DataType.FLOAT16
    if dtype == trt.bfloat16:
        return TensorProto.DataType.BFLOAT16
    elif dtype == trt.float32:
        return TensorProto.DataType.FLOAT
    elif dtype == trt.int32:
        return TensorProto.DataType.INT32
    elif dtype == trt.int64:
        return TensorProto.DataType.INT64
    elif dtype == trt.fp8:
        return TensorProto.DataType.FLOAT8E4M3FN
    else:
        raise TypeError("%s is not supported" % dtype)


def to_onnx(network, path: Path):
    inputs = []
    for i in range(network.num_inputs):
        network_input = network.get_input(i)
        inputs.append(
            helper.make_tensor_value_info(
                network_input.name,
                trt_dtype_to_onnx(network_input.dtype),
                list(network_input.shape),
            )
        )

    outputs = []
    for i in range(network.num_outputs):
        network_output = network.get_output(i)
        outputs.append(
            helper.make_tensor_value_info(
                network_output.name,
                trt_dtype_to_onnx(network_output.dtype),
                list(network_output.shape),
            )
        )

    nodes = []
    for i in range(network.num_layers):
        layer = network.get_layer(i)
        layer_inputs = []
        for j in range(layer.num_inputs):
            ipt = layer.get_input(j)
            if ipt is not None:
                layer_inputs.append(layer.get_input(j).name)
        layer_outputs = [layer.get_output(j).name for j in range(layer.num_outputs)]
        nodes.append(
            helper.make_node(
                str(layer.type),
                name=layer.name,
                inputs=layer_inputs,
                outputs=layer_outputs,
                domain="com.nvidia",
            )
        )

    onnx_model = helper.make_model(
        helper.make_graph(nodes, "attention", inputs, outputs, initializer=None),
        producer_name="NVIDIA",
    )
    onnx.save(onnx_model, path)



import pytest
import torch

from optimum.nvidia.quantization import Float8QuantizationConfig
from optimum.nvidia.quantization.ammo.config import KV_CACHE_CFG


@pytest.mark.parametrize("kv_cache", (False, True))
@pytest.mark.parametrize("lm_head", (False, True))
def test_float8_ammo_configuration(kv_cache, lm_head):
    config = Float8QuantizationConfig(
        with_quantized_kv_cache=kv_cache, with_quantized_lm_head=lm_head
    )

    assert config.requires_calibration
    assert not config.has_calibration_dataset
    assert config.weight_dtype == torch.float8_e4m3fn

    assert config.has_quantized_lm_head == lm_head
    assert config.has_quantized_kv_cache == kv_cache

    qconfig = config.as_ammo_config()
    assert qconfig is not None
    assert "quant_cfg" in qconfig

    quant_cfg = qconfig["quant_cfg"]

    for pattern in {"*weight_quantizer", "*input_quantizer", "default"}:
        assert pattern in quant_cfg
        assert quant_cfg[pattern]["num_bits"] == (4, 3)
        assert quant_cfg[pattern]["axis"] is None

    assert quant_cfg["*lm_head*"]["enable"] == lm_head

    if kv_cache:
        for kv_pattern in KV_CACHE_CFG.keys():
            assert quant_cfg[kv_pattern]["enable"]
            assert quant_cfg[kv_pattern]["axis"] is None
            assert quant_cfg[kv_pattern]["num_bits"] == (4, 3)



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
import os

from transformers.utils import strtobool


INT_TRUE_VALUE = 1


@functools.cache
def parse_flag_from_env(name: str, default: bool) -> bool:
    """
    Parse the environment variable `name` as a boolean
    :param name: Name of target environment variable
    :param default: The default value to apply if `name` is not present
    :return: Boolean value
    """

    # Retrieve the value or `default` if not present
    value = os.environ.get(name, str(default))

    try:
        return strtobool(value) == INT_TRUE_VALUE
    except ValueError:
        raise ValueError(
            f"Failed to convert environment variable {name}={value} to a bool"
        )



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import tensorrt_llm.bindings as ctrrt
import torch
from transformers import GenerationConfig


if TYPE_CHECKING:
    from transformers import PretrainedConfig

LOGGER = getLogger(__name__)

PackedTensor = List[torch.Tensor]


DEFAULT_BATCH_SIZE: int = 1
DEFAULT_PROMPT_LENGTH: int = 128
DEFAULT_BEAM_WIDTH: int = 1


class CompiledModel:
    def __init__(self, engines_folders_path: List[Union[Path, PathLike]]):
        # A compiled model may have several subfolders (e.g. encoder-decoder model).
        self._engines_folders_path = [
            Path(engines_folder_path) for engines_folder_path in engines_folders_path
        ]

    @property
    def engine_path(self) -> Path:
        """
        Return the local path where the engine(s) is/are located
        :return: Path to the folder holding the engine(s) definition(s)
        """
        return self._engines_folders_path


class CausalLM(CompiledModel):
    __slots__ = (
        "_device",
        "_config",
        "_mapping",
        "_session",
        "_session_config",
        "_use_packed_inputs",
        "max_beam_width",
        "max_batch_size",
        "max_prompt_length",
        "max_output_length",
    )

    def __init__(
        self,
        engines_folders: List[Path],
        *,
        gpus_per_node: int,
        transformers_config: "PretrainedConfig",
        use_cuda_graph: bool = False,
        generation_config: Optional[GenerationConfig] = None,
    ):
        if len(engines_folders) != 1:
            raise ValueError(
                f"For CausalLM, expecting a single engine folder, got: {engines_folders}"
            )
        super().__init__(engines_folders)
        engines_folder = engines_folders[0]

        self._device = torch.device("cuda")
        self._config = ctrrt.GptJsonConfig.parse_file(engines_folder / "config.json")
        self._mapping = ctrrt.WorldConfig.mpi(
            gpus_per_node,
            self._config.tensor_parallelism,
            self._config.pipeline_parallelism,
        )
        self._session_config = ctrrt.GptSessionConfig(
            max_batch_size=self._config.model_config.max_batch_size,
            max_beam_width=self._config.model_config.max_beam_width,
            max_sequence_length=self._config.model_config.max_seq_len,
        )
        self._session_config.cuda_graph_mode = use_cuda_graph

        # Create the engine
        engine_file = self._config.engine_filename(self._mapping)
        self._session = ctrrt.GptSession(
            config=self._session_config,
            model_config=self._config.model_config,
            world_config=self._mapping,
            engine_file=str(engines_folder.joinpath(engine_file)),
        )

        # Additional cached properties
        self._use_packed_inputs = self._config.model_config.use_packed_input
        self.max_batch_size = self._config.model_config.max_batch_size
        self.max_prompt_length = self._config.model_config.max_input_len

        self.max_output_length = self._config.model_config.max_seq_len
        self.max_beam_width = self._session_config.max_beam_width

        if generation_config is None:
            generation_config = GenerationConfig()
        self.generation_config = generation_config

        self.transformers_config = transformers_config

    @property
    def config(self) -> ctrrt.GptJsonConfig:
        return self._config

    def generate(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        max_new_tokens: int = -1,
        min_length: int = -1,
        num_beams: int = 1,
        temperature: float = 1.0,
        top_k: int = 50,
        top_p: float = 1.0,
        repetition_penalty: float = 1.0,
        length_penalty: float = 1.0,
        seed: int = 0,
        pad_token_id: int = 0,
        bos_token_id: int = 1,
        eos_token_id: int = 2,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        device = self._device

        # If no GenerationConfig is provided, let's allocate one with default settings
        generation_config = ctrrt.SamplingConfig(min(num_beams, self.max_beam_width))
        generation_config.random_seed = [seed]
        generation_config.temperature = [temperature]
        generation_config.top_k = [top_k]
        generation_config.top_p = [top_p]
        generation_config.repetition_penalty = [repetition_penalty]
        generation_config.length_penalty = [length_penalty]

        if min_length > 0:
            generation_config.min_length = [min_length]

        with torch.no_grad():
            if not isinstance(input_ids, torch.Tensor):
                raise TypeError("input_ids should be a PyTorch tensor (torch.Tensor)")

            input_ids, lengths = self._prepare_inputs(input_ids, attention_mask)
            if torch.any(torch.gt(lengths, self.max_prompt_length)):
                raise ValueError(
                    f"Input length {lengths} is bigger than maximum prompt length ({self.max_prompt_length})."
                )

            trt_inputs = ctrrt.GenerationInput(
                end_id=eos_token_id,
                pad_id=pad_token_id,
                ids=input_ids.to(device),
                lengths=lengths.to(device),
                packed=self._use_packed_inputs,
            )

            if max_new_tokens is None or max_new_tokens < 1:
                max_new_tokens = self.max_output_length - input_ids.shape[1]

            trt_inputs.max_new_tokens = max_new_tokens

            # Tensors are being allocated as in/out parameters and TRTLLM will resize
            trt_outputs = ctrrt.GenerationOutput(
                ids=torch.empty(0, device=device, dtype=torch.int32),
                lengths=torch.empty(0, device=device, dtype=torch.int32),
            )

            self._session.generate(trt_outputs, trt_inputs, generation_config)

            return trt_outputs.ids, trt_outputs.lengths

    def _prepare_inputs(
        self, input_ids: torch.Tensor, attention_mask: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        shape = input_ids.size()
        input_ids = input_ids.int()

        if input_ids.ndim == 1:
            lengths = torch.tensor(shape, dtype=torch.int32, device=self._device)
        elif input_ids.ndim == 2 and shape[0] == 1:
            lengths = torch.tensor([shape[1]], dtype=torch.int32, device=self._device)
        elif attention_mask is not None:
            lengths = attention_mask.sum(dim=1, dtype=torch.int32)
        else:
            warnings.warn(
                "Not enough information to compute the non-padded tensor length. "
                "Please provide an attention_mask to avoid situations where padding"
                " will be attended to in attention modules"
            )

            attention_mask = torch.ones_like(input_ids)
            lengths = torch.tensor(shape, dtype=torch.int32).flatten()

        if self._use_packed_inputs and shape[0] > 1:
            input_ids = torch.masked_select(input_ids, attention_mask.bool()).view(
                1, -1
            )

        return input_ids, lengths


class TensorRTForSpeechSeq2Seq(CompiledModel):
    def __init__(
        self,
        engines_folders: List[Path],
        *,
        gpus_per_node: int,
        transformers_config: "PretrainedConfig",
        use_cuda_graph: bool = False,
        generation_config: Optional[GenerationConfig] = None,
    ):
        super().__init__(engines_folders)



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import Optional, Tuple

import pytest
from tensorrt_llm.quantization import QuantMode

from optimum.nvidia.config import convert_quant_method_to_trt


@pytest.mark.parametrize(
    "method, weight_num_bits, activation_num_bits, expected",
    (
        (
            "awq",
            4,
            None,
            (
                QuantMode.from_description(
                    quantize_weights=True, per_group=True, use_int4_weights=True
                ),
                "W4A16_AWQ",
            ),
        ),
        (
            "awq",
            8,
            None,
            (
                QuantMode.from_description(
                    quantize_weights=True, per_group=True, use_int4_weights=False
                ),
                "W8A16_AWQ",
            ),
        ),
        (
            "gptq",
            4,
            None,
            (
                QuantMode.from_description(
                    quantize_weights=True, per_group=False, use_int4_weights=True
                ),
                "W4A16_GPTQ",
            ),
        ),
        (
            "gptq",
            8,
            None,
            (
                QuantMode.from_description(
                    quantize_weights=True, per_group=False, use_int4_weights=False
                ),
                "W8A16_GPTQ",
            ),
        ),
    ),
)
def test_convert_quantization_config_to_trt(
    method: str,
    weight_num_bits: int,
    activation_num_bits: Optional[int],
    expected: Tuple,
):
    assert (
        convert_quant_method_to_trt(method, weight_num_bits, activation_num_bits)
        == expected
    )



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import Optional, Protocol, runtime_checkable

from tensorrt_llm import Mapping
from tensorrt_llm.quantization import QuantMode


@runtime_checkable
class SupportsFromHuggingFace(Protocol):
    """
    Define the protocol implemented by TensorRT-LLM models to support loading from Hugging Face Hub
    """

    @classmethod
    def from_hugging_face(
        cls,
        hf_model_dir,
        dtype="float16",
        mapping: Optional[Mapping] = None,
        quant_mode: Optional[QuantMode] = None,
        **kwargs,
    ):
        """

        :param hf_model_dir:
        :param dtype:
        :param mapping:
        :param quant_mode:
        :param kwargs:
        :return:
        """
        ...



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Union

import torch
from tensorrt_llm import Mapping
from tensorrt_llm.models import PretrainedConfig as TensorRTPretrainedConfig
from tensorrt_llm.models.modeling_utils import (
    QuantizationConfig as TensorRTQuantizationConfig,
)
from tensorrt_llm.plugin import PluginConfig
from tensorrt_llm.quantization import QuantMode
from transformers import AutoConfig, PretrainedConfig

from optimum.nvidia.quantization import AmmoQuantizationConfig
from optimum.nvidia.utils import get_user_agent


SUPPORTED_DTYPES = [
    "float32",
    "float16",
    "bfloat16",
    "int32",
    "int8",
    "uint8",
    "float8",
]


def dtype_to_str(dtype: Union[torch.dtype, str]) -> str:
    if isinstance(dtype, str):
        if dtype in SUPPORTED_DTYPES:
            return dtype
        else:
            raise ValueError(f"Unsupported dtype ({dtype}) value")
    elif dtype == torch.float32:
        return "float32"
    elif dtype == torch.float16:
        return "float16"
    elif dtype == torch.bfloat16:
        return "bfloat16"
    elif dtype == torch.int32:
        return "int32"
    elif dtype == torch.int8:
        return "int8"
    elif dtype == torch.uint8:
        return "uint8"
    elif dtype == torch.float8_e4m3fn:
        return "float8"
    else:
        raise ValueError(f"Unsupported torch.dtype ({dtype}) value")


def convert_quant_method_to_trt(
    method: str, weight_num_bits: int, activation_num_bits: Optional[int] = None
) -> (QuantMode, str):
    if method == "awq":
        if not activation_num_bits:
            activation_num_bits = 16

        if weight_num_bits not in {4, 8}:
            raise ValueError(
                f"Unsupported AWQ quantization schema with {weight_num_bits}-bits weights. "
                "Only 4 and 8-bits weights' quantization schemas are supported."
            )

        if activation_num_bits not in {8, 16}:
            raise ValueError(
                f"Unsupported AWQ quantization schema with {activation_num_bits}-bits activations. "
                "Only 8 and 16-bits activations' quantization schemas are supported."
            )

        mode = QuantMode.from_description(
            quantize_weights=True, per_group=True, use_int4_weights=weight_num_bits == 4
        )
        return mode, f"W{weight_num_bits}A{activation_num_bits}_AWQ"
    elif method == "gptq":
        if not activation_num_bits:
            activation_num_bits = 16

        if weight_num_bits not in {4, 8}:
            raise ValueError(
                f"Unsupported GPTQ quantization schema with {weight_num_bits}-bits weights. "
                "Only 4 and 8-bits weights' quantization schemas are supported."
            )

        if activation_num_bits != 16:
            raise ValueError(
                f"Unsupported GPTQ quantization schema with {activation_num_bits}-bits activations. "
                "Only 16-bits activations' quantization schemas are supported."
            )

        mode = QuantMode.from_description(
            quantize_weights=True,
            per_group=False,
            use_int4_weights=weight_num_bits == 4,
        )
        return mode, f"W{weight_num_bits}A16_GPTQ"
    else:
        raise ValueError(f"Unsupported quantization method: {method}")


class TensorRTConfig(ABC, TensorRTPretrainedConfig):
    @staticmethod
    def get_quantization_config(
        config: PretrainedConfig,
    ) -> (QuantMode, AmmoQuantizationConfig):
        if hasattr(config, "quantization_config"):
            qconfig = config.quantization_config
            num_bits = qconfig.num_bits
            group_size = qconfig.group_size
            mode, quant_method = convert_quant_method_to_trt(
                qconfig.quant_method, num_bits
            )
            has_zero_point = qconfig.get("zero_point", False)
            exclude_modules = qconfig.get("module_to_not_convert", [])

            return mode, TensorRTQuantizationConfig(
                quantization_algo=quant_method,
                kv_cache_quant_algo=None,
                group_size=group_size,
                has_zero_point=has_zero_point,
                exclude_modules=exclude_modules,
            )
        else:
            return QuantMode.from_description(), TensorRTQuantizationConfig(
                None, None, None, False, None
            )

    @staticmethod
    @abstractmethod
    def from_config(config: PretrainedConfig) -> "TensorRTConfig":
        raise NotImplementedError()

    @staticmethod
    def from_pretrained(
        model_id_or_path: str,
        revision: Optional[str] = None,
        token: Union[bool, str, None] = None,
    ):
        config = AutoConfig.from_pretrained(
            model_id_or_path,
            revision=revision,
            token=token,
            user_agent=get_user_agent(),
        )
        return TensorRTConfig.from_config(config)

    @staticmethod
    @abstractmethod
    def supports_strong_typing() -> bool:
        raise NotImplementedError()

    def shard(
        self,
        world_size: int,
        gpus_per_node: int,
        rank: int = 0,
        tp_degree: int = 1,
        pp_degree: int = 1,
    ):
        if tp_degree * pp_degree != world_size:
            raise ValueError(
                f"tensor parallelism ({tp_degree}) x pipeline parallelism ({pp_degree})"
                f" != world size ({world_size})"
            )

        self.mapping = Mapping(
            world_size=world_size,
            rank=rank,
            gpus_per_node=gpus_per_node,
            tp_size=tp_degree,
            pp_size=pp_degree,
        )

    def get_plugins_config(self) -> PluginConfig:
        return PluginConfig(
            rmsnorm_quantization_plugin="disable",
            layernorm_quantization_plugin="disable",
            nccl_plugin="disable",
            paged_kv_cache="enable",
            enable_xqa="enable",
            use_paged_context_fmha=None,
            use_context_fmha_for_generation=None,
            tokens_per_block=None,
            attention_qk_half_accumulation=None,
            multi_block_mode=None,
            use_custom_all_reduce=True,
            remove_input_padding=True,
            context_fmha=None,
            context_fmha_fp32_acc=None,
        )

    def save_pretrained(self, save_folder: Union[str, Path]):
        config_dict = self.to_dict()
        config_dict.pop("trt_model_class", None)
        config_dict.pop("trt_model_file", None)

        with open(Path(save_folder, "config.json"), "w") as config_f:
            json.dump(config_dict, config_f)



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest

from optimum.nvidia.utils import parse_flag_from_env
from optimum.nvidia.utils.constants import SM_ADA_LOVELACE
from optimum.nvidia.utils.nvml import get_device_compute_capabilities, get_device_count


# Environment variable controlling test set
ENVVAR_NAME_RUN_NIGHTLY = "RUN_NIGHTLY"
ENVVAR_NAME_RUN_SLOW = "RUN_SLOW"
ENVVAR_NAME_RUN_CPU_ONLY = "RUN_CPU_ONLY"


nightly = pytest.mark.skipif(
    parse_flag_from_env(ENVVAR_NAME_RUN_NIGHTLY, False), reason="Nightly test"
)
slow = pytest.mark.skipif(
    parse_flag_from_env(ENVVAR_NAME_RUN_SLOW, False), reason="Slow test"
)

requires_gpu = pytest.mark.skipif(
    parse_flag_from_env(ENVVAR_NAME_RUN_CPU_ONLY, False) or not get_device_count(),
    reason=f"RUN_CPU_ONLY={parse_flag_from_env(ENVVAR_NAME_RUN_CPU_ONLY, False)} or "
    f"no GPU detected (num_gpus={get_device_count()})",
)


def requires_gpu_compute_capabilities_ge(min_capabilities: int, device: int = 0):
    (major, minor) = get_device_compute_capabilities(device)
    compute_capabilities = major * 10 + minor
    return pytest.mark.skipif(
        compute_capabilities < min_capabilities,
        reason=f"Require compute capabilities >= sm_{min_capabilities} but current GPU is sm_{compute_capabilities}",
    )


requires_float8 = requires_gpu_compute_capabilities_ge(SM_ADA_LOVELACE)



#  coding=utf-8
#  Copyright 2023 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import copy
import json
import pathlib
from collections import OrderedDict
from logging import getLogger
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import torch
from tensorrt_llm import Mapping, mpi_rank, str_dtype_to_torch
from tensorrt_llm._utils import str_dtype_to_trt, torch_to_numpy, trt_dtype_to_torch
from tensorrt_llm.builder import BuildConfig
from tensorrt_llm.functional import LayerNormPositionType, LayerNormType
from tensorrt_llm.models import DecoderModel as TrtDecoderModel
from tensorrt_llm.models import PretrainedConfig, PretrainedModel
from tensorrt_llm.models import WhisperEncoder as TrtWhisperEncoder
from tensorrt_llm.plugin import PluginConfig
from tensorrt_llm.runtime import GenerationSession, ModelConfig, SamplingConfig
from tensorrt_llm.runtime.generation import LogitsProcessorList
from tensorrt_llm.runtime.session import Session, TensorInfo
from transformers import GenerationConfig
from transformers import PreTrainedModel as TransformersPretrainedModel
from transformers.models.whisper.generation_whisper import WhisperGenerationMixin
from transformers.models.whisper.modeling_whisper import (
    WhisperDecoder as TransformersWhisperDecoder,
)
from transformers.models.whisper.modeling_whisper import (
    WhisperEncoder as TransformersWhisperEncoder,
)
from transformers.models.whisper.modeling_whisper import (
    WhisperForConditionalGeneration as TransformersWhisperForConditionalGeneration,
)
from transformers.models.whisper.tokenization_whisper import TASK_IDS, TO_LANGUAGE_CODE

from optimum.nvidia import TensorRTConfig
from optimum.nvidia.config import dtype_to_str
from optimum.nvidia.generation.logits_process import (
    TrtForceTokensLogitsProcessor,
    TrtSuppressTokensAtBeginLogitsProcessor,
    TrtSuppressTokensLogitsProcessor,
    TrtWhisperNoSpeechDetection,
)
from optimum.nvidia.hub import HuggingFaceHubModel
from optimum.nvidia.runtime import TensorRTForSpeechSeq2Seq
from optimum.nvidia.utils.nvml import get_max_memory


if TYPE_CHECKING:
    from transformers import PretrainedConfig as TransformersPretrainedConfig
    from transformers.generation.logits_process import LogitsProcessorList
    from transformers.generation.stopping_criteria import StoppingCriteriaList


LOGGER = getLogger(__name__)


def split(v, tp_size, idx, dim=0):
    if tp_size == 1:
        return v
    return np.split(v, tp_size, axis=dim)[idx]


def get_qkv(model_params, torch_dtype, attn_prefix: str):
    q_weight = torch_to_numpy(
        model_params[f"{attn_prefix}.q_proj.weight"].to(torch_dtype)
    )
    k_weight = torch_to_numpy(
        model_params[f"{attn_prefix}.k_proj.weight"].to(torch_dtype)
    )
    v_weight = torch_to_numpy(
        model_params[f"{attn_prefix}.v_proj.weight"].to(torch_dtype)
    )

    qkv_weight = (q_weight, k_weight, v_weight)

    # At least one of the query, key, value projection has bias.
    if any(
        bias_name in model_params
        for bias_name in [
            f"{attn_prefix}.q_proj.bias",
            f"{attn_prefix}.k_proj.bias",
            f"{attn_prefix}.v_proj.bias",
        ]
    ):
        # For example whisper encoder k_proj does not have bias, so we will just fill with zeros the fused bias if needed.
        numpy_precision = q_weight.dtype
        if f"{attn_prefix}.q_proj.bias" in model_params:
            q_bias = torch_to_numpy(
                model_params[f"{attn_prefix}.q_proj.bias"].to(torch_dtype)
            )
        else:
            q_bias = np.zeros(q_weight.shape[0], dtype=numpy_precision)

        if f"{attn_prefix}.k_proj.bias" in model_params:
            k_bias = torch_to_numpy(
                model_params[f"{attn_prefix}.k_proj.bias"].to(torch_dtype)
            )
        else:
            k_bias = np.zeros(k_weight.shape[0], dtype=numpy_precision)

        if f"{attn_prefix}.v_proj.bias" in model_params:
            v_bias = torch_to_numpy(
                model_params[f"{attn_prefix}.v_proj.bias"].to(torch_dtype)
            )
        else:
            v_bias = np.zeros(v_weight.shape[0], dtype=numpy_precision)
        qkv_bias = (q_bias, k_bias, v_bias)
    else:
        qkv_bias = None

    return qkv_weight, qkv_bias


def convert_from_hf_whisper_encoder(
    hf_whisper_encoder,
    mapping=Mapping(),
    dtype="float16",
):
    num_layers = hf_whisper_encoder.config.encoder_layers
    torch_dtype = str_dtype_to_torch(dtype)

    model_params = dict(hf_whisper_encoder.named_parameters())
    weights = {}

    # Convert specific tensors
    # conv1
    # TensorRT-LLM Conv1d weight is 4D while transformers checkpoint ones are 3D.
    conv1_weight = torch_to_numpy(model_params["conv1.weight"].to(torch_dtype))
    conv1_weight = conv1_weight[..., None]

    weights["model.conv1.weight"] = conv1_weight
    weights["model.conv1.bias"] = torch_to_numpy(
        model_params["conv1.bias"].to(torch_dtype)
    )

    # conv2
    conv2_weight = torch_to_numpy(model_params["conv2.weight"].to(torch_dtype))
    conv2_weight = conv2_weight[..., None]

    weights["model.conv2.weight"] = conv2_weight
    weights["model.conv2.bias"] = torch_to_numpy(
        model_params["conv2.bias"].to(torch_dtype)
    )

    # embed_positions
    # NOTE: this one is kept as fp32 in Whisper, is this important?
    weights["model.positional_embedding"] = torch_to_numpy(
        model_params["embed_positions.weight"]  # .to(torch_dtype)
    )

    # Final layer norm
    weights["model.ln_post.weight"] = torch_to_numpy(
        model_params["layer_norm.weight"].to(torch_dtype)
    )
    weights["model.ln_post.bias"] = torch_to_numpy(
        model_params["layer_norm.bias"].to(torch_dtype)
    )

    # Map all the hidden layers
    for layer_idx in range(num_layers):
        prefix = f"layers.{layer_idx}"

        # attention_layernorm
        weights[f"model.encoder_layers.{layer_idx}.attention_layernorm.weight"] = (
            torch_to_numpy(
                model_params[f"{prefix}.self_attn_layer_norm.weight"].to(torch_dtype)
            )
        )
        weights[f"model.encoder_layers.{layer_idx}.attention_layernorm.bias"] = (
            torch_to_numpy(
                model_params[f"{prefix}.self_attn_layer_norm.bias"].to(torch_dtype)
            )
        )

        # mlp_layernorm
        weights[f"model.encoder_layers.{layer_idx}.mlp_layernorm.weight"] = (
            torch_to_numpy(
                model_params[f"{prefix}.final_layer_norm.weight"].to(torch_dtype)
            )
        )
        weights[f"model.encoder_layers.{layer_idx}.mlp_layernorm.bias"] = (
            torch_to_numpy(
                model_params[f"{prefix}.final_layer_norm.bias"].to(torch_dtype)
            )
        )

        # Self attention layer
        # TensorRT-LLM model definition uses a single GEMM for query/key/value, while transformers does not.
        qkv_weight, qkv_bias = get_qkv(
            model_params, attn_prefix=f"{prefix}.self_attn", torch_dtype=torch_dtype
        )
        q_weight, k_weight, v_weight = qkv_weight

        qkv_weight = np.concatenate((q_weight, k_weight, v_weight), axis=0)
        weight = split(qkv_weight, mapping.tp_size, mapping.tp_rank, dim=1)
        weights[f"model.encoder_layers.{layer_idx}.attention.qkv.weight"] = weight

        if qkv_bias is not None:
            q_bias, k_bias, v_bias = qkv_bias
            packed_qkv_bias = np.concatenate((q_bias, k_bias, v_bias), axis=0)
            weights[f"model.encoder_layers.{layer_idx}.attention.qkv.bias"] = (
                np.ascontiguousarray(packed_qkv_bias)
            )

        # Common projection logic
        # 0: column tensor parallel, 1: row tensor parallel.
        for src, dst, shard_axis in [
            ("self_attn.out_proj.weight", "attention.dense.weight", 1),
            ("fc1.weight", "mlp.fc.weight", 0),
            ("fc2.weight", "mlp.proj.weight", 1),
        ]:
            weight = torch_to_numpy(model_params[f"{prefix}.{src}"].to(torch_dtype))
            weight = split(weight, mapping.tp_size, mapping.tp_rank, dim=shard_axis)
            weights[f"model.encoder_layers.{layer_idx}.{dst}"] = weight

        # Bias is never sharded.
        for src, dst in [
            ("self_attn.out_proj.bias", "attention.dense.bias"),
            ("fc1.bias", "mlp.fc.bias"),
            ("fc2.bias", "mlp.proj.bias"),
        ]:
            weights[f"model.encoder_layers.{layer_idx}.{dst}"] = torch_to_numpy(
                model_params[f"{prefix}.{src}"].to(torch_dtype)
            )

    # weights["lm_head.weight"] = np.zeros((0))  # Just a hack for commands/build.py

    return weights


def convert_from_hf_whisper_decoder(
    hf_whisper_decoder,
    mapping=Mapping(),
    dtype="float16",
):
    weights = {}

    num_layers = hf_whisper_decoder.config.decoder_layers
    torch_dtype = str_dtype_to_torch(dtype)

    model_params = dict(hf_whisper_decoder.named_parameters())

    # Convert specific tensors
    if mapping.is_first_pp_rank():
        # embed_tokens
        weights["model.embedding.vocab_embedding.weight"] = torch_to_numpy(
            model_params["embed_tokens.weight"].to(torch_dtype)
        )

        # embed_positions
        weights["model.embedding.position_embedding.weight"] = torch_to_numpy(
            model_params["embed_positions.weight"].to(torch_dtype)
        )

    if mapping.is_last_pp_rank():
        # Final layer norm
        weights["model.final_layernorm.weight"] = torch_to_numpy(
            model_params["layer_norm.weight"].to(torch_dtype)
        )
        weights["model.final_layernorm.bias"] = torch_to_numpy(
            model_params["layer_norm.bias"].to(torch_dtype)
        )

        # Final vocab projection
        lm_head = torch_to_numpy(model_params["embed_tokens.weight"].to(torch_dtype))
        lm_head = split(lm_head, mapping.tp_size, mapping.tp_rank)
        weights["model.lm_head.weight"] = lm_head

    # Map all the hidden layers
    for layer_idx in range(num_layers):
        prefix = f"layers.{layer_idx}"
        trt_llm_prefix = f"model.decoder_layers.{layer_idx}"

        # self_attention_layernorm
        weights[f"{trt_llm_prefix}.self_attention_layernorm.weight"] = torch_to_numpy(
            model_params[f"{prefix}.self_attn_layer_norm.weight"].to(torch_dtype)
        )
        weights[f"{trt_llm_prefix}.self_attention_layernorm.bias"] = torch_to_numpy(
            model_params[f"{prefix}.self_attn_layer_norm.bias"].to(torch_dtype)
        )

        # cross_attention_layernorm
        weights[f"{trt_llm_prefix}.cross_attention_layernorm.weight"] = torch_to_numpy(
            model_params[f"{prefix}.encoder_attn_layer_norm.weight"].to(torch_dtype)
        )
        weights[f"{trt_llm_prefix}.cross_attention_layernorm.bias"] = torch_to_numpy(
            model_params[f"{prefix}.encoder_attn_layer_norm.bias"].to(torch_dtype)
        )

        # mlp_layernorm
        weights[f"{trt_llm_prefix}.mlp_layernorm.weight"] = torch_to_numpy(
            model_params[f"{prefix}.final_layer_norm.weight"].to(torch_dtype)
        )
        weights[f"{trt_llm_prefix}.mlp_layernorm.bias"] = torch_to_numpy(
            model_params[f"{prefix}.final_layer_norm.bias"].to(torch_dtype)
        )

        # Self attention layer
        qkv_weight, qkv_bias = get_qkv(
            model_params, attn_prefix=f"{prefix}.self_attn", torch_dtype=torch_dtype
        )
        q_weight, k_weight, v_weight = qkv_weight

        qkv_weight = np.concatenate((q_weight, k_weight, v_weight), axis=0)
        weight = split(qkv_weight, mapping.tp_size, mapping.tp_rank, dim=1)
        weights[f"{trt_llm_prefix}.self_attention.qkv.weight"] = weight

        if qkv_bias is not None:
            q_bias, k_bias, v_bias = qkv_bias
            packed_qkv_bias = np.concatenate((q_bias, k_bias, v_bias), axis=0)
            weights[f"{trt_llm_prefix}.self_attention.qkv.bias"] = np.ascontiguousarray(
                packed_qkv_bias
            )

        # Cross attention layer
        qkv_weight, qkv_bias = get_qkv(
            model_params, attn_prefix=f"{prefix}.encoder_attn", torch_dtype=torch_dtype
        )
        q_weight, k_weight, v_weight = qkv_weight

        qkv_weight = np.concatenate((q_weight, k_weight, v_weight), axis=0)
        weight = split(qkv_weight, mapping.tp_size, mapping.tp_rank, dim=1)
        weights[f"{trt_llm_prefix}.cross_attention.qkv.weight"] = weight

        if qkv_bias is not None:
            q_bias, k_bias, v_bias = qkv_bias
            packed_qkv_bias = np.concatenate((q_bias, k_bias, v_bias), axis=0)
            weights[f"{trt_llm_prefix}.cross_attention.qkv.bias"] = (
                np.ascontiguousarray(packed_qkv_bias)
            )

        # Common projection logic.
        # 0: column tensor parallel, 1: row tensor parallel.
        for src, dst, shard_axis in [
            ("self_attn.out_proj.weight", "self_attention.dense.weight", 1),
            ("encoder_attn.out_proj.weight", "cross_attention.dense.weight", 1),
            ("fc1.weight", "mlp.fc.weight", 0),
            ("fc2.weight", "mlp.proj.weight", 1),
        ]:
            weight = torch_to_numpy(model_params[f"{prefix}.{src}"].to(torch_dtype))
            weight = split(weight, mapping.tp_size, mapping.tp_rank, dim=shard_axis)
            weights[f"{trt_llm_prefix}.{dst}"] = weight

        # Bias is never sharded.
        for src, dst in [
            ("self_attn.out_proj.bias", "self_attention.dense.bias"),
            ("encoder_attn.out_proj.bias", "cross_attention.dense.bias"),
            ("fc1.bias", "mlp.fc.bias"),
            ("fc2.bias", "mlp.proj.bias"),
        ]:
            weights[f"{trt_llm_prefix}.{dst}"] = torch_to_numpy(
                model_params[f"{prefix}.{src}"].to(torch_dtype)
            )

    return weights


class WhisperEncoderConfig(TensorRTConfig):
    @classmethod
    def from_config(cls, config: "TransformersPretrainedConfig") -> "TensorRTConfig":
        # Retrieve the quantization from the transformers config (if provided)
        _, qconfig = TensorRTConfig.get_quantization_config(config)

        trt_config = cls(
            architecture=config.architectures[0],
            dtype=dtype_to_str(config.torch_dtype),  # TODO: always float32?
            logits_dtype="float32",
            vocab_size=config.vocab_size,
            max_position_embeddings=config.max_target_positions,
            hidden_size=config.d_model,
            num_hidden_layers=config.encoder_layers,
            num_attention_heads=config.encoder_attention_heads,
            num_key_value_heads=config.encoder_attention_heads,
            hidden_act=config.activation_function,
            intermediate_size=None,
            norm_epsilon=None,
            position_embedding_type="learned_absolute",
            world_size=1,
            tp_size=1,
            pp_size=1,
            quantization=qconfig,
            use_parallel_embedding=None,
            embedding_sharding_dim=None,
            share_embedding_table=None,
            head_size=-1,  # We need to set it otherwise TRT-LLM tries to compute `hidden_size // num_attention_heads`
            max_source_positions=config.max_source_positions,
            num_mel_bins=config.num_mel_bins,
            trt_model_class="TrtWhisperEncoderPretrainedModel",
            trt_model_file=pathlib.Path(__file__),
        )

        trt_config.mapping.gpus_per_node = min(trt_config.mapping.world_size, 8)

        return trt_config

    def get_plugins_config(self) -> PluginConfig:
        config = super().get_plugins_config()
        config.bert_attention_plugin = self.dtype
        config.gpt_attention_plugin = "disable"
        config.remove_input_padding = False  # This one is bugged with Whisper.
        config.paged_kv_cache = "disable"  # TODO: getting AssertionError: Paged kv cache is enabled, the kv_cache_block_pointers tensor shall not be None

        config.moe_plugin = "disable"
        config.gemm_plugin = self.dtype
        config.context_fmha = True
        config.enable_xqa = False
        config.remove_input_padding = False
        config.use_custom_all_reduce = "disable"

        config.layernorm_quantization_plugin = None
        config.rmsnorm_quantization_plugin = None
        config.nccl_plugin = None

        return config

    @staticmethod
    def supports_strong_typing() -> bool:
        return False


class WhisperDecoderConfig(TensorRTConfig):
    @classmethod
    def from_config(cls, config: "TransformersPretrainedConfig") -> "TensorRTConfig":
        # Retrieve the quantization from the transformers config (if provided)
        _, qconfig = TensorRTConfig.get_quantization_config(config)

        trt_config = cls(
            architecture=config.architectures[0],
            dtype=dtype_to_str(config.torch_dtype),  # TODO: always float32?
            logits_dtype="float32",
            vocab_size=config.vocab_size,
            max_position_embeddings=config.max_target_positions,
            hidden_size=config.d_model,
            num_hidden_layers=config.decoder_layers,
            num_attention_heads=config.decoder_attention_heads,
            num_key_value_heads=config.decoder_attention_heads,
            hidden_act=config.activation_function,
            intermediate_size=None,
            norm_epsilon=None,
            position_embedding_type="learned_absolute",
            world_size=1,
            tp_size=1,
            pp_size=1,
            quantization=qconfig,
            use_parallel_embedding=None,
            embedding_sharding_dim=None,
            share_embedding_table=None,
            head_size=-1,  # We need to set it otherwise TRT-LLM tries to compute `hidden_size // num_attention_heads`
            max_source_positions=config.max_source_positions,
            decoder_ffn_dim=config.decoder_ffn_dim,
            trt_model_class="TrtWhisperDecoderPretrainedModel",
            trt_model_file=pathlib.Path(__file__),
            num_encoder_attention_heads=config.encoder_attention_heads,
        )

        trt_config.mapping.gpus_per_node = min(trt_config.mapping.world_size, 8)

        return trt_config

    def get_plugins_config(self) -> PluginConfig:
        config = super().get_plugins_config()
        config.bert_attention_plugin = "disable"
        config.gpt_attention_plugin = self.dtype
        config.paged_kv_cache = "disable"  # TODO: getting AssertionError: Paged kv cache is enabled, the kv_cache_block_pointers tensor shall not be None

        config.context_fmha = True
        config.moe_plugin = "disable"
        config.gemm_plugin = self.dtype
        config.remove_input_padding = False
        config.enable_xqa = False

        config.layernorm_quantization_plugin = None
        config.rmsnorm_quantization_plugin = None
        config.nccl_plugin = None
        config.use_custom_all_reduce = "disable"

        return config

    @staticmethod
    def supports_strong_typing() -> bool:
        return False


class TrtWhisperEncoderPretrainedModel(PretrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.model = TrtWhisperEncoder(
            n_mels=config.num_mel_bins,
            n_ctx=config.max_source_positions,
            n_state=config.hidden_size,
            n_head=config.num_attention_heads,
            n_layer=config.num_hidden_layers,
            dtype=str_dtype_to_trt(config.dtype),
        )

    def forward(self, **kwargs):
        return self.model(**kwargs)

    def prepare_inputs(self, max_batch_size=16, **kwargs):
        (x, input_lengths) = self.model.prepare_inputs(max_batch_size=max_batch_size)

        return {"x": x, "input_lengths": input_lengths}


class TrtWhisperDecoderPretrainedModel(PretrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.model = TrtDecoderModel(
            num_layers=config.num_hidden_layers,
            num_heads=config.num_attention_heads,
            hidden_size=config.hidden_size,
            ffn_hidden_size=config.decoder_ffn_dim,
            encoder_num_heads=config.num_encoder_attention_heads,
            encoder_hidden_size=config.hidden_size,
            vocab_size=config.vocab_size,
            dtype=str_dtype_to_trt(config.dtype),
            logits_dtype=str_dtype_to_trt(config.logits_dtype),
            max_position_embeddings=config.max_position_embeddings,
            has_position_embedding=True,
            relative_attention=False,
            head_size=None,
            encoder_head_size=None,
            num_kv_heads=None,
            encoder_num_kv_heads=None,
            type_vocab_size=None,
            max_distance=0,
            num_buckets=0,
            has_embedding_layernorm=False,
            has_embedding_scale=False,
            q_scaling=1.0,
            has_attention_qkvo_bias=True,
            has_mlp_bias=True,
            has_model_final_layernorm=True,
            layernorm_eps=1e-5,
            layernorm_position=LayerNormPositionType.pre_layernorm,
            layernorm_type=LayerNormType.LayerNorm,
            hidden_act=config.hidden_act,
            rescale_before_lm_head=False,
            mapping=config.mapping,
        )
        self.config.optimize_network = False  # See utils/patching.py. TODO: remove this once native to TensorRT-LLM.

    def forward(self, **kwargs):
        return self.model(**kwargs)

    def prepare_inputs(
        self,
        max_batch_size,
        max_input_len,
        max_seq_len,
        use_cache,
        max_beam_width: int = 1,
        max_num_tokens: int = None,
        prompt_embedding_table_size: int = 0,
        position_encoding_2d: bool = False,
        max_draft_len: int = 0,
        gather_context_logits: bool = False,
        gather_generation_logits: bool = False,
        lora_target_modules: List[str] = None,
    ):
        if not use_cache:
            raise NotImplementedError("use_cache=False is not implemented for Whisper.")

        (
            input_ids,
            encoder_output,
            position_ids,
            token_type_ids,
            use_cache,
            attention_mask,
            cross_attention_mask,
            last_token_ids,
            kv_cache_params,
            attention_params,
            hidden_states,
            lora_params,
            cross_kv_cache_gen,
            cross_qkv_reuse,
        ) = self.model.prepare_inputs(
            max_batch_size=max_batch_size,
            max_beam_width=max_beam_width,
            max_decoder_input_len=max_input_len,
            max_new_tokens=max_seq_len,
            max_encoder_input_len=self.config.max_source_positions,
        )

        return {
            "decoder_input_ids": input_ids,
            "encoder_output": encoder_output,
            "position_ids": position_ids,
            "token_type_ids": token_type_ids,
            "use_cache": use_cache,
            "attention_mask": attention_mask,
            "cross_attention_mask": cross_attention_mask,
            "last_token_ids": last_token_ids,
            "kv_cache_params": kv_cache_params,
            "attention_params": attention_params,
            "hidden_states": hidden_states,
            "lora_params": lora_params,
            "cross_kv_cache_gen": cross_kv_cache_gen,
            "cross_qkv_reuse": cross_qkv_reuse,
        }


class OptimumWhisperEncoder(TensorRTForSpeechSeq2Seq, HuggingFaceHubModel):
    MODEL_CONFIG = WhisperEncoderConfig
    HF_LIBRARY_TARGET_MODEL_CLASS = TransformersWhisperEncoder
    TRT_LLM_TARGET_MODEL_CLASS = TrtWhisperEncoderPretrainedModel

    @staticmethod
    def convert_weights(
        target: PretrainedModel,
        source: TransformersPretrainedModel,
        config: PretrainedConfig,
    ) -> Dict[str, torch.Tensor]:
        if config.quant_mode.has_any_quant():
            raise NotImplementedError("Quantization is not supported yet.")

        return convert_from_hf_whisper_encoder(source, config.mapping, config.dtype)


class OptimumWhisperDecoder(TensorRTForSpeechSeq2Seq, HuggingFaceHubModel):
    MODEL_CONFIG = WhisperDecoderConfig
    HF_LIBRARY_TARGET_MODEL_CLASS = TransformersWhisperDecoder
    TRT_LLM_TARGET_MODEL_CLASS = TrtWhisperDecoderPretrainedModel

    @staticmethod
    def convert_weights(
        target: PretrainedModel,
        source: TransformersPretrainedModel,
        config: PretrainedConfig,
    ) -> Dict[str, torch.Tensor]:
        if config.quant_mode.has_any_quant():
            raise NotImplementedError("Quantization is not supported yet.")

        return convert_from_hf_whisper_decoder(source, config.mapping, config.dtype)


class WhisperForConditionalGeneration(
    TensorRTForSpeechSeq2Seq, HuggingFaceHubModel, WhisperGenerationMixin
):
    MODEL_CONFIG = None
    HF_LIBRARY_TARGET_MODEL_CLASS = TransformersWhisperForConditionalGeneration
    TRT_LLM_TARGET_MODEL_CLASS = None  # Whisper is split in two in TRT-LLM.

    def __init__(
        self,
        engines_folders: List[Path],
        *,
        gpus_per_node: int,
        transformers_config: "TransformersPretrainedConfig",
        use_cuda_graph: bool = False,
        generation_config: Optional[GenerationConfig] = None,
    ):
        super().__init__(
            engines_folders,
            gpus_per_node=gpus_per_node,
            transformers_config=transformers_config,
            use_cuda_graph=use_cuda_graph,
            generation_config=generation_config,
        )

        if generation_config is None:
            generation_config = GenerationConfig()
        self.generation_config = generation_config

        self.transformers_config = transformers_config

        # Encoder.
        serialize_path = engines_folders[0] / "rank0.engine"
        with open(serialize_path, "rb") as f:
            encoder_session = Session.from_serialized_engine(f.read())

        self.encoder_session = encoder_session

        # Decoder.
        decoder_config_path = engines_folders[1] / "config.json"
        with open(decoder_config_path, "r") as f:
            decoder_config = json.load(f)

        serialize_path = engines_folders[1] / "rank0.engine"
        with open(serialize_path, "rb") as f:
            decoder_engine_buffer = f.read()

        build_config = BuildConfig.from_dict(decoder_config["build_config"])
        trt_config = WhisperDecoderConfig.from_dict(decoder_config["pretrained_config"])

        self.dtype = trt_config.dtype

        decoder_model_config = ModelConfig(
            max_batch_size=build_config.max_batch_size,
            max_beam_width=build_config.max_beam_width,
            num_heads=trt_config.num_attention_heads,
            num_kv_heads=trt_config.num_key_value_heads,
            hidden_size=trt_config.hidden_size,
            vocab_size=trt_config.vocab_size,
            num_layers=trt_config.num_hidden_layers,
            gpt_attention_plugin=build_config.plugin_config.gpt_attention_plugin,
            remove_input_padding=build_config.plugin_config.remove_input_padding,
            cross_attention=True,
            has_position_embedding=True,
            has_token_type_embedding=False,
        )

        # world_size > 1 is not supported.
        world_size = 1
        runtime_rank = mpi_rank()

        runtime_mapping = Mapping(world_size, runtime_rank)

        decoder_generation_session = GenerationSession(
            decoder_model_config,
            decoder_engine_buffer,
            runtime_mapping,
            debug_mode=False,
        )

        self.decoder_generation_session = decoder_generation_session

    @classmethod
    def convert_and_build(
        cls,
        local_path: Path,
        hf_model_config: Dict,
        engine_save_path: Optional[Path] = None,
        hf_model: Optional[TransformersPretrainedModel] = None,
        **model_kwargs,
    ) -> Path:
        max_memory = get_max_memory()

        if hf_model is None:
            # Allocate required components for quantization
            hf_model = cls.HF_LIBRARY_TARGET_MODEL_CLASS.from_pretrained(
                local_path,
                device_map="auto",
                max_memory=max_memory,
                local_files_only=True,
            )

        if engine_save_path is None:
            engine_save_path = local_path

        LOGGER.info("Building Whisper encoder...")
        (
            encoder_engines_folder,
            encoder_engines_relative_folder,
        ) = OptimumWhisperEncoder.convert_and_build(
            local_path,
            hf_model_config,
            engine_save_path=Path(engine_save_path, "encoder"),
            hf_model=hf_model.model.encoder,
            config_class=WhisperEncoderConfig,
            **model_kwargs,
        )

        LOGGER.info("Building Whisper decoder...")
        (
            decoder_engines_folder,
            decoder_engines_relative_folder,
        ) = OptimumWhisperDecoder.convert_and_build(
            local_path,
            hf_model_config,
            engine_save_path=Path(engine_save_path, "decoder"),
            hf_model=hf_model.model.decoder,
            config_class=WhisperDecoderConfig,
            **model_kwargs,
        )

        return [encoder_engines_folder[0], decoder_engines_folder[0]], [
            encoder_engines_relative_folder[0],
            decoder_engines_relative_folder[0],
        ]

    def encoder(
        self,
        input_features: torch.Tensor,
    ):
        if dtype_to_str(input_features.dtype) != self.dtype:
            LOGGER.warning(
                f"input_features should be of dtype {self.dtype}, got {dtype_to_str(input_features.dtype)}. Automatically casting to {self.dtype}."
            )
            input_features = input_features.to(str_dtype_to_torch(self.dtype))

        input_lengths = torch.tensor(
            [input_features.shape[2] // 2 for _ in range(input_features.shape[0])],
            dtype=torch.int32,
            device=input_features.device,
        )

        inputs = OrderedDict()
        inputs["x"] = input_features
        inputs["input_lengths"] = input_lengths

        output_list = [
            TensorInfo("x", str_dtype_to_trt(self.dtype), input_features.shape),
            TensorInfo("input_lengths", str_dtype_to_trt("int32"), input_lengths.shape),
        ]

        output_info = (self.encoder_session).infer_shapes(output_list)

        LOGGER.debug(f"output info {output_info}")
        outputs = {
            t.name: torch.empty(
                tuple(t.shape), dtype=trt_dtype_to_torch(t.dtype), device="cuda"
            )
            for t in output_info
        }

        stream = torch.cuda.current_stream()

        ok = self.encoder_session.run(
            inputs=inputs, outputs=outputs, stream=stream.cuda_stream
        )

        assert ok, "Engine execution failed"
        stream.synchronize()

        return outputs["output"]

    def _retrieve_logit_processors(
        self, generation_config, logits_processor, begin_index, is_shortform, num_beams
    ):
        # Adapted from WhisperGenerationMixin._retrieve_logit_processors with XxxLogitsProcessor -> TrtXxxLogitsProcessor

        if generation_config.return_timestamps is True:
            # TODO: implement.
            raise NotImplementedError(
                "return_timestamps=True is not implemented with TensorRT-LLM. Please open an issue at https://github.com/huggingface/optimum-nvidia/issues. In the meanwhile, please set `model.generation_config.return_timestamps=False`."
            )

        if generation_config.suppress_tokens is not None:
            suppress_tokens_processor = TrtSuppressTokensLogitsProcessor(
                generation_config.suppress_tokens
            )
            logits_processor = (
                [suppress_tokens_processor]
                if logits_processor is None
                else [suppress_tokens_processor] + logits_processor
            )
            generation_config.suppress_tokens = None

        if generation_config.begin_suppress_tokens is not None:
            begin_suppress_processor = TrtSuppressTokensAtBeginLogitsProcessor(
                generation_config.begin_suppress_tokens, begin_index=begin_index
            )
            logits_processor = (
                [begin_suppress_processor]
                if logits_processor is None
                else [begin_suppress_processor] + logits_processor
            )
            generation_config.begin_suppress_tokens = None

        if generation_config.no_speech_threshold is not None and not is_shortform:
            no_speech_detector = TrtWhisperNoSpeechDetection(
                no_speech_token=generation_config.no_timestamps_token_id - 1,
                begin_index=begin_index,
                scores_is_logprobs=num_beams > 1,
            )
            logits_processor = (
                [no_speech_detector]
                if logits_processor is None
                else [no_speech_detector] + logits_processor
            )
            no_speech_detector.set_model(self)

        if is_shortform and generation_config.forced_decoder_ids is not None:
            forced_tokens_proc = TrtForceTokensLogitsProcessor(
                generation_config.forced_decoder_ids
            )
            # It's important that the `forced_tokens_proc` processor is appended after
            # the suppress_tokens processor or else it might happen that all token logits are suppressed to -inf
            # which would lead to unexpected behavior
            # The better approach here is to NOT make use of the `forced_tokens_proc` for Whisper and instead
            # initialize all of them as `decoder_input_ids`.
            # TODO(Sanchit): Make sure to deprecate this in v4.39 as there will be no `forced_decoder_ids` anymore.
            logits_processor = (
                [forced_tokens_proc]
                if logits_processor is None
                else logits_processor + [forced_tokens_proc]
            )
            generation_config.forced_decoder_ids = None

        return logits_processor

    def _retrieve_init_tokens(
        self, input_features, generation_config, config, num_segment_frames, kwargs
    ):
        # Adapted from WhisperGenerationMixin._retrieve_init_tokens with automatic language detection disabled.

        def replace_or_add(lst: List[int], num: int, itr: Iterator[int]):
            """short function to replace num with a itr in lst"""
            found = any(i in lst for i in itr)
            if found:
                lst = [num if i in itr else i for i in lst]
            else:
                lst.append(num)
            return lst

        task = getattr(generation_config, "task", None)
        language = getattr(generation_config, "language", None)

        if kwargs.get("forced_decoder_ids", None) is not None:
            forced_decoder_ids = kwargs["forced_decoder_ids"]
        elif (
            hasattr(generation_config, "forced_decoder_ids")
            and generation_config.forced_decoder_ids is not None
        ):
            forced_decoder_ids = generation_config.forced_decoder_ids

            if language is None and task is None and forced_decoder_ids[0][1] is None:
                LOGGER.warning_once(
                    "Due to a bug fix in https://github.com/huggingface/transformers/pull/28687 transcription using a multilingual Whisper will default to language detection followed by transcription instead of translation to English."
                    "This might be a breaking change for your use case. If you want to instead always translate your audio to English, make sure to pass `language='en'`."
                )
        elif (
            hasattr(config, "forced_decoder_ids")
            and config.forced_decoder_ids is not None
        ):
            forced_decoder_ids = config.forced_decoder_ids
        else:
            forced_decoder_ids = None

        if forced_decoder_ids is not None and task is not None:
            LOGGER.info(
                f"You have passed task={task}, but also have set `forced_decoder_ids` to {forced_decoder_ids} which creates a conflict. `forced_decoder_ids` will be ignored in favor of task={task}."
            )
            forced_decoder_ids = None
        elif forced_decoder_ids is not None and language is not None:
            LOGGER.info(
                f"You have passed language={language}, but also have set `forced_decoder_ids` to {forced_decoder_ids} which creates a conflict. `forced_decoder_ids` will be ignored in favor of language={language}."
            )
            forced_decoder_ids = None

        init_tokens = [generation_config.decoder_start_token_id]
        if forced_decoder_ids is not None and forced_decoder_ids[0][0] == 1:
            i = 1
            while len(forced_decoder_ids) > 0 and forced_decoder_ids[0][0] == i:
                init_tokens += [forced_decoder_ids[0][1]]
                forced_decoder_ids = forced_decoder_ids[1:]
                i += 1

            if len(forced_decoder_ids) > 0:
                raise ValueError(
                    f"You are using token ids in `forced_decoder_ids` that do not seem to correctly follow the prompt pattern of Whisper. Make sure that {forced_decoder_ids} has an entry for all indices >= 1 and < {forced_decoder_ids[0][0]}.",
                )

        # from v4.39 the forced decoder ids are always None in favour of decoder input ids
        generation_config.forced_decoder_ids = None

        is_lang_id_undefined = len(init_tokens) <= 1 or (
            len(init_tokens) > 1 and init_tokens[1] is None
        )
        if language is not None:
            if language in generation_config.lang_to_id.keys():
                language_token = language
            elif language in TO_LANGUAGE_CODE.keys():
                language_token = f"<|{TO_LANGUAGE_CODE[language]}|>"
            elif language in TO_LANGUAGE_CODE.values():
                language_token = f"<|{language}|>"
            else:
                is_language_code = len(language) == 2
                raise ValueError(
                    f"Unsupported language: {language}. Language should be one of:"
                    f" {list(TO_LANGUAGE_CODE.values()) if is_language_code else list(TO_LANGUAGE_CODE.keys())}."
                )
            if language_token not in generation_config.lang_to_id:
                raise ValueError(
                    f"{language_token} is not supported by this specific model as it is not in the `generation_config.lang_to_id`."
                    "(You should just add it to the generation config)"
                )

            lang_id = generation_config.lang_to_id[language_token]

            # if language is defined it'll overwrite language ids that might have already been defined via the generation_config
            replace_or_add(init_tokens, lang_id, generation_config.lang_to_id.values())
        elif hasattr(generation_config, "lang_to_id") and is_lang_id_undefined:
            raise ValueError(
                f"The language is not specified in the model's generation_config, and automatic language detection is not supported with TensorRT-LLM. Please set e.g. model.generation_config.language = '<|en|>' for English language. Available languages: {generation_config.lang_to_id.keys()}. Please refer to https://github.com/huggingface/transformers/blob/v4.39.3/src/transformers/models/whisper/tokenization_whisper.py#L95 for the languages codes."
            )

        if task is not None:
            if task in TASK_IDS:
                init_tokens.append(generation_config.task_to_id[generation_config.task])
                task_id = generation_config.task_to_id[generation_config.task]

                # if task is defined it'll overwrite task ids that might have already been defined via the generation_config
                replace_or_add(
                    init_tokens, task_id, generation_config.task_to_id.values()
                )
            else:
                raise ValueError(
                    f"The `{task}`task is not supported. The task should be one of `{TASK_IDS}`"
                )
        elif language is not None and hasattr(generation_config, "task_to_id"):
            # if language is defined, but no task id is in `init_tokens`, default to transcribe
            if not any(i in init_tokens for i in generation_config.task_to_id.values()):
                init_tokens.append(generation_config.task_to_id["transcribe"])

        if (
            not generation_config.return_timestamps
            and hasattr(generation_config, "no_timestamps_token_id")
            and init_tokens[-1] != generation_config.no_timestamps_token_id
        ):
            init_tokens.append(generation_config.no_timestamps_token_id)
        elif (
            generation_config.return_timestamps
            and init_tokens[-1] == generation_config.no_timestamps_token_id
        ):
            LOGGER.info(
                "<|notimestamps|> prompt token is removed from generation_config since `return_timestamps` is set to `'True'`."
            )
            init_tokens = init_tokens[:-1]

        # let's make sure we don't pass `None` tokens as prompt tokens
        init_tokens = [t for t in init_tokens if t is not None]

        return init_tokens

    def generate(
        self,
        inputs: Optional[torch.Tensor] = None,
        generation_config: Optional[GenerationConfig] = None,
        logits_processor: Optional["LogitsProcessorList"] = None,
        stopping_criteria: Optional["StoppingCriteriaList"] = None,
        prefix_allowed_tokens_fn: Optional[
            Callable[[int, torch.Tensor], List[int]]
        ] = None,
        synced_gpus: bool = False,
        return_timestamps: Optional[bool] = None,
        task: Optional[str] = None,
        language: Optional[str] = None,
        is_multilingual: Optional[bool] = None,
        prompt_ids: Optional[torch.Tensor] = None,
        prompt_condition_type: Optional[str] = None,  # first-segment, all-segments
        condition_on_prev_tokens: Optional[bool] = None,
        temperature: Optional[Union[float, Tuple[float, ...]]] = None,
        compression_ratio_threshold: Optional[float] = None,
        logprob_threshold: Optional[float] = None,
        no_speech_threshold: Optional[float] = None,
        num_segment_frames: Optional[int] = None,
        attention_mask: Optional[torch.Tensor] = None,
        time_precision: float = 0.02,
        return_token_timestamps: Optional[bool] = None,
        return_segments: bool = False,
        return_dict_in_generate: Optional[bool] = None,
        **kwargs,
    ):
        if inputs.device.type != "cuda":
            raise ValueError(
                f"TensorRT-LLM only supports inputs on CUDA device. Got: inputs.device = {inputs.device}"
            )

        def raise_unsupported(value: Any, name: str, default: Any = None):
            if value != default:
                raise ValueError(
                    f"TensorRTForSpeechSeq2Seq.generate does not support {name} (got {value}). Please open an issue at https://github.com/huggingface/optimum-nvidia/issues."
                )

        raise_unsupported(stopping_criteria, name="stopping_criteria")
        raise_unsupported(prefix_allowed_tokens_fn, name="prefix_allowed_tokens_fn")
        raise_unsupported(synced_gpus, name="synced_gpus", default=False)
        raise_unsupported(return_timestamps, name="return_timestamps")
        raise_unsupported(task, name="task")
        raise_unsupported(prompt_ids, name="prompt_ids")
        raise_unsupported(prompt_condition_type, name="prompt_condition_type")
        raise_unsupported(temperature, name="temperature")
        raise_unsupported(attention_mask, name="attention_mask")
        raise_unsupported(time_precision, name="time_precision", default=0.02)
        raise_unsupported(return_token_timestamps, name="return_token_timestamps")
        raise_unsupported(return_segments, name="return_segments", default=False)
        raise_unsupported(return_dict_in_generate, name="return_dict_in_generate")

        # 1. copy generation config
        if generation_config is None:
            generation_config = copy.deepcopy(self.generation_config)
        else:
            generation_config = copy.deepcopy(generation_config)

        self._set_language_and_task(
            language=language,
            task=task,
            is_multilingual=is_multilingual,
            generation_config=generation_config,
        )
        self._set_token_ids(
            generation_config=generation_config,
            config=self.transformers_config,
            kwargs=kwargs,
        )
        self._set_thresholds_and_condition(
            generation_config=generation_config,
            logprob_threshold=logprob_threshold,
            compression_ratio_threshold=compression_ratio_threshold,
            no_speech_threshold=no_speech_threshold,
            condition_on_prev_tokens=condition_on_prev_tokens,
        )

        num_beams = kwargs.pop("num_beams", generation_config.num_beams)
        input_stride = 1 * 2  # encoder's conv1 stride * encoder's conv2 stride

        batch_size, total_input_frames = self._retrieve_total_input_frames(
            input_features=inputs, input_stride=input_stride, kwargs=kwargs
        )
        num_segment_frames = (
            input_stride * self.transformers_config.max_source_positions
        )
        is_shortform = total_input_frames <= num_segment_frames
        if not is_shortform:
            raise ValueError(
                "Whisper TensorRT-LLM implementation only supports short form for now. Please open an issue at https://github.com/huggingface/optimum-nvidia/issues."
            )

        init_tokens = self._retrieve_init_tokens(
            inputs,
            generation_config=generation_config,
            config=self.transformers_config,
            num_segment_frames=num_segment_frames,
            kwargs=kwargs,
        )

        begin_index = len(init_tokens)
        logits_processor = self._retrieve_logit_processors(
            generation_config=generation_config,
            logits_processor=logits_processor,
            begin_index=begin_index,  # begin index is index of first generated decoder token
            is_shortform=is_shortform,
            num_beams=kwargs.get("num_beams", 1),
        )
        logits_processor = LogitsProcessorList(logits_processor)

        encoder_outputs = self.encoder(inputs)

        batch_size = inputs.shape[0]
        one_tensor = torch.ones((batch_size, 1), device="cuda", dtype=torch.long)
        decoder_input_ids = torch.cat([t * one_tensor for t in init_tokens], dim=-1)

        max_new_tokens = kwargs.pop("max_new_tokens", generation_config.max_new_tokens)
        if max_new_tokens is None:
            # Transformers' GenerationConfig.max_new_tokens defaults to None.
            if generation_config.max_length is not None:
                max_new_tokens = (
                    generation_config.max_length - decoder_input_ids.shape[1]
                )
            else:
                raise ValueError("Please specifiy the argument `max_new_tokens`.")

        if (
            max_new_tokens + decoder_input_ids.shape[-1]
            > self.transformers_config.max_target_positions
        ):
            max_new_tokens = kwargs.get("max_new_tokens", 0)
            raise ValueError(
                f"The length of `decoder_input_ids` equal `prompt_ids` plus special start tokens is {decoder_input_ids.shape[-1]}, and the `max_new_tokens` "
                f"is {max_new_tokens}. Thus, the combined length of "
                f"`decoder_input_ids` and `max_new_tokens` is: {max_new_tokens + decoder_input_ids.shape[-1]}. This exceeds the "
                f"`max_target_positions` of the Whisper model: {self.transformers_config.max_target_positions}. "
                "You should either reduce the length of your prompt, or reduce the value of `max_new_tokens`, "
                f"so that their combined length is less than {self.transformers_config.max_target_positions}."
            )

        encoder_input_lengths = torch.tensor(
            [encoder_outputs.shape[1] for x in range(encoder_outputs.shape[0])],
            dtype=torch.int32,
            device=inputs.device,
        )

        decoder_input_lengths = torch.tensor(
            [decoder_input_ids.shape[-1] for _ in range(decoder_input_ids.shape[0])],
            dtype=torch.int32,
            device="cuda",
        )
        decoder_max_input_length = torch.max(decoder_input_lengths).item()

        cross_attention_mask = torch.ones(
            [encoder_outputs.shape[0], 1, encoder_outputs.shape[1]],
            device=inputs.device,
            dtype=torch.int32,
        )

        sampling_config = SamplingConfig(
            end_id=generation_config.eos_token_id,
            pad_id=generation_config.pad_token_id,
            num_beams=num_beams,
        )

        self.decoder_generation_session.setup(
            decoder_input_lengths.size(0),
            decoder_max_input_length,
            max_new_tokens,
            beam_width=num_beams,
            encoder_max_input_length=encoder_outputs.shape[1],
        )

        torch.cuda.synchronize()

        decoder_input_ids = decoder_input_ids.type(torch.int32).cuda()

        # output_ids of shape [batch_size, beam_width, output_len]
        output_ids = self.decoder_generation_session.decode(
            decoder_input_ids,
            decoder_input_lengths,
            sampling_config,
            encoder_output=encoder_outputs,
            encoder_input_lengths=encoder_input_lengths,
            cross_attention_mask=cross_attention_mask,
            logits_processor=logits_processor,
        )
        torch.cuda.synchronize()

        return output_ids[
            :,
            0,
            : torch.max(self.decoder_generation_session.sequence_length_buffer) + 1,
        ]



from pathlib import Path
from tempfile import TemporaryDirectory

import mock
from transformers import AutoConfig as HfAutoConfig
from transformers import AutoModelForCausalLM as HfAutoModelForCausalLM

# import pytest
import optimum.nvidia.hub
from optimum.nvidia import AutoModelForCausalLM
from optimum.nvidia.hub import FOLDER_TRTLLM_ENGINES


# from optimum.nvidia.utils.nvml import get_device_name


# def test_load_engine_from_huggingface_hub():
#     with mock.patch("optimum.nvidia.hub.HuggingFaceHubModel.convert_and_build"):
#         device = get_device_name(0)
#
#         try:
#             model = AutoModelForCausalLM.from_pretrained(
#                 "optimum-nvidia/llama-ci", revision=device[-1].lower()
#             )
#             assert model is not None
#             assert (
#                 optimum.nvidia.hub.HuggingFaceHubModel.convert_and_build.call_count == 0
#             )
#         except ValueError:
#             pytest.skip(
#                 f"No revision found for optimum-nvidia/llama-ci on GPU: {device[-1].lower()}"
#             )


def test_save_engine_locally_and_reload():
    with TemporaryDirectory() as out:
        out = Path(out)
        hf_out = out.joinpath("_hf")
        trtllm_out = out.joinpath("_trtllm")

        def _save():
            config = HfAutoConfig.from_pretrained("meta-llama/Llama-2-7b-chat-hf")
            config.num_hidden_layers = 1

            model = HfAutoModelForCausalLM.from_config(config)
            model.save_pretrained(hf_out)

            model = AutoModelForCausalLM.from_pretrained(hf_out)
            model.save_pretrained(trtllm_out)

            assert trtllm_out.exists()
            assert (trtllm_out / FOLDER_TRTLLM_ENGINES).exists()
            assert (trtllm_out / FOLDER_TRTLLM_ENGINES / "rank0.engine").exists()

        def _reload():
            with mock.patch("optimum.nvidia.hub.HuggingFaceHubModel.convert_and_build"):
                model = AutoModelForCausalLM.from_pretrained(trtllm_out)
                assert model is not None
                assert (
                    optimum.nvidia.hub.HuggingFaceHubModel.convert_and_build.call_count
                    == 0
                )

        _save()
        _reload()



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from optimum.nvidia.utils.nvml import SM_FP8_SUPPORTED


class OptimumNvidiaException(Exception):
    def __init__(self, operation: str, msg: str):
        super().__init__(f"[{operation}] {msg}.")


### Model support
class UnsupportedModelException(OptimumNvidiaException):
    def __init__(self, model_type: str):
        super(
            f"Model of type {model_type} is not supported. "
            "Please open-up an issue at https://github.com/huggingface/optimum-nvidia/issues"
        )


### Unsupported features blocks
class UnsupportedHardwareFeature(OptimumNvidiaException):
    """
    Base exception class for all features not supported by underlying hardware
    """

    def __init__(self, msg):
        super("feature", msg)

    @classmethod
    def float8(cls) -> "UnsupportedHardwareFeature":
        return Float8NotSupported()


class Float8NotSupported(UnsupportedHardwareFeature):
    """
    Thrown when attempting to target float8 inference but the underlying hardware doesn't support it
    """

    def __init__(self):
        super(
            "float8 is not supported on your device. "
            f"Please use a device with compute capabilities {SM_FP8_SUPPORTED}"
        )



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import shutil
from glob import glob, iglob
from logging import getLogger
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
    runtime_checkable,
)
from warnings import warn

import numpy as np
import torch
from huggingface_hub import ModelHubMixin, snapshot_download
from huggingface_hub.hub_mixin import T
from safetensors.torch import save_file as to_safetensors
from tensorrt_llm._utils import numpy_to_torch
from tensorrt_llm.models.modeling_utils import PretrainedConfig, PretrainedModel
from transformers import AutoConfig, AutoTokenizer, GenerationConfig
from transformers import PreTrainedModel as TransformersPretrainedModel
from transformers.utils import SAFE_WEIGHTS_INDEX_NAME

from optimum.nvidia import TensorRTConfig
from optimum.nvidia.builder import LocalEngineBuilder
from optimum.nvidia.builder.config import EngineConfigBuilder
from optimum.nvidia.quantization import AutoQuantizationConfig
from optimum.nvidia.quantization.ammo import AmmoQuantizer
from optimum.nvidia.utils import get_user_agent, maybe_offload_weights_to_cpu
from optimum.nvidia.utils.nvml import get_max_memory


ATTR_TRTLLM_ENGINE_FOLDER = "__trtllm_engine_folder__"
FOLDER_TRTLLM_ENGINES = "engines"
FILE_TRTLLM_ENGINE_PATTERN = "rank[0-9]*.engine"

HUB_TRTLLM_ENGINE_PATTERNS = ["**/config.json", f"**/{FILE_TRTLLM_ENGINE_PATTERN}"]
HUB_SAFETENSORS_PATTERNS = ["config.json", "*.safetensors", SAFE_WEIGHTS_INDEX_NAME]
LOGGER = getLogger()
LOGGER.setLevel(level="INFO")


def extract_model_type(config: Dict[str, Any]) -> Tuple[Optional[str], bool]:
    if "model_type" in config:
        model_type = config["model_type"]
        is_tensorrt_config = False

    # This path try to extract from the TensorRTLLM config
    elif (
        "pretrained_config" in config and "architecture" in config["pretrained_config"]
    ):
        model_type = config["pretrained_config"]["architecture"]
        prefix_pos = model_type.index("For")
        model_type = model_type[:prefix_pos].lower()
        is_tensorrt_config = True
    else:
        return None, False

    return model_type, is_tensorrt_config


def find_prebuilt_engines(root: Path) -> Tuple[List[Path], List[Path]]:
    """
    Attempt to locate any prebuilt TRT engines at the provided root path or root' subfolders.
    :param root: The directory we should look into for the engine files.
    :return: The list of `Path` where engines were found in root (or root/FOLDER_TRTLLM_ENGINES, or root/**/FOLDER_TRTLLM_ENGINES), and the corresponding relative path to root.
    """
    folders = []
    relative_folders = []

    files = glob(Path(root, f"**/{FOLDER_TRTLLM_ENGINES}/rank*.engine").as_posix())
    for file in files:
        file_directory = Path(Path(file).parents[0])
        folders.append(file_directory)
        relative_folders.append(file_directory.relative_to(root))

    files = glob(Path(root, "rank*.engine").as_posix())
    if len(files) > 0:
        folders.append(root)
        relative_folders.append(".")

    files = glob(Path(root, f"{FOLDER_TRTLLM_ENGINES}/rank*.engine").as_posix())
    if len(files) > 0:
        folders.append(Path(root, FOLDER_TRTLLM_ENGINES))
        relative_folders.append(FOLDER_TRTLLM_ENGINES)

    folders = list(dict.fromkeys(folders))
    relative_folders = list(dict.fromkeys(relative_folders))

    # TODO: Handle this properly - we should enforce the directory names and use dicts instead of lists here.
    if len(relative_folders) == 2 and "decoder" in relative_folders[0].as_posix():
        folders.reverse()
        relative_folders.reverse()

    return folders, relative_folders


@runtime_checkable
class SupportsTensorrtConversion(Protocol):
    MODEL_CONFIG: Type[TensorRTConfig]
    HF_LIBRARY_TARGET_MODEL_CLASS: Type[ModelHubMixin]
    TRT_LLM_TARGET_MODEL_CLASS: Type[PretrainedModel]

    @staticmethod
    def convert_weights(
        target: PretrainedModel,
        source: TransformersPretrainedModel,
        config: PretrainedConfig,
    ) -> Dict[str, np.ndarray]: ...


class HuggingFaceHubModel(ModelHubMixin, SupportsTensorrtConversion):
    @classmethod
    def convert_and_build(
        cls,
        local_path: Path,
        hf_model_config: Dict,
        engine_save_path: Optional[Path] = None,
        hf_model: Optional[TransformersPretrainedModel] = None,
        config_class: Optional[TensorRTConfig] = None,
        **model_kwargs,
    ) -> Tuple[List[Path], List[Path]]:
        """
        :param local_path:
        :param hf_model_config:
        :param model_kwargs:
        :return:
        """

        # Path where will be stored the engines
        if engine_save_path is None:
            engine_save_path = local_path
            engines_folder = local_path / FOLDER_TRTLLM_ENGINES
            engines_folder.mkdir(exist_ok=True, parents=True)
        else:
            engines_folder = engine_save_path / FOLDER_TRTLLM_ENGINES
            engines_folder.mkdir(exist_ok=True, parents=True)

        # Retrieve configuration
        config = AutoConfig.for_model(**hf_model_config)
        if "torch_dtype" in model_kwargs:
            config.torch_dtype = model_kwargs["torch_dtype"]

        # Convert the original config to a model config TRTLLM understands
        model_config = HuggingFaceHubModel.convert_config_to_trtllm(
            cls, config, config_class=config_class, **model_kwargs
        )

        # We now have a TRTLLM compatible config, so let's feed it to the target TRTLLM model to create a checkpoint
        LOGGER.debug("Allocating TRTLLM model to build the checkpoint")
        model = cls.TRT_LLM_TARGET_MODEL_CLASS.from_config(model_config)

        # Retrieve the parameters for building the engine
        if "engine_config" in model_kwargs:
            engine_config = model_kwargs.pop("engine_config")
        else:
            builder = EngineConfigBuilder.from_dict(model_config, **model_kwargs)
            builder.with_plugins_config(model_config.get_plugins_config())
            engine_config = builder.build()

        if engine_config.plugins_config is None:
            engine_config.plugins_config = model_config.get_plugins_config()

        max_memory = get_max_memory()

        if hf_model is None:
            LOGGER.debug(
                f"Loading weights from {local_path} into the model ({cls.HF_LIBRARY_TARGET_MODEL_CLASS.__name__})"
            )

            hf_model = cls.HF_LIBRARY_TARGET_MODEL_CLASS.from_pretrained(
                local_path,
                device_map="auto",
                max_memory=max_memory,
                local_files_only=True,
            )
        else:
            if not isinstance(hf_model, cls.HF_LIBRARY_TARGET_MODEL_CLASS):
                raise ValueError(
                    f"Expected a {cls.HF_LIBRARY_TARGET_MODEL_CLASS.__name__} model to be provided, but the argument `hf_model` is a {hf_model.__class__.__name__}."
                )

        hf_model = hf_model.eval()
        hf_model = maybe_offload_weights_to_cpu(hf_model)

        # Retrieve potential quantization config (If provided) - follow the transformers parameter's name
        has_qconfig = "quantization_config" in model_kwargs
        use_fp8 = model_kwargs.get("use_fp8", False)

        if has_qconfig or use_fp8:
            LOGGER.debug("About to quantize Hugging Face model")

            if has_qconfig:
                qconfig = model_kwargs.pop("quantization_config")
            elif use_fp8:
                if (
                    candidate_tokenizer_path := engines_folder.parent.joinpath(
                        "tokenizer.json"
                    )
                ).exists():
                    tokenizer_path = candidate_tokenizer_path.parent
                elif "_model_id" in hf_model_config:
                    tokenizer_path = hf_model_config["_model_id"]
                else:
                    raise ValueError(
                        "Unable to determine the tokenizer to use to quantize this model. "
                        "Please provide a complete QuantizationConfig using "
                        "from_pretrained(..., quantization_config=AutoQuantizationConfig.from_description())"
                    )

                tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
                qconfig = AutoQuantizationConfig.from_description(
                    weight="float8",
                    activation="float8",
                    tokenizer=tokenizer,
                    dataset="c4-new",
                )

                warn(
                    "Converting model to support float8 inference.\n"
                    f"Calibrating model with dataset='c4', split='train', samples={len(qconfig.calibration_dataset)}.\n"
                    "Note: if text generation doesn't meet your expectations, "
                    "you can control the quantization process manually with this API: "
                    "qconfig = AutoQuantizationConfig.from_description(weight='float8', activation='float8', ...) "
                    "forwarding the configuration to .from_pretrained(..., quantization_config=qconfig)"
                )

            hf_quantizer = AmmoQuantizer(
                quantization_config=qconfig,
                artifact_path=engines_folder,
                tensor_parallel_degree=engine_config.sharding_profile.tensor_parallelism,
                pipeline_parallel_degree=engine_config.sharding_profile.pipeline_parallelism,
                export_tensorrt_llm_config=True,
            )

            hf_quantizer.preprocess_model(hf_model, batch_size=1)
            hf_quantizer.postprocess_model(hf_model)

        else:
            # Apply the conversion from Hugging Face weights to TRTLLM
            for rank in range(model_config.mapping.world_size):
                LOGGER.debug(
                    f"Converting weights from Hugging Face checkpoint for rank {rank}"
                )
                model_config.set_rank(rank)
                converted_weights = cls.convert_weights(model, hf_model, model_config)
                converted_weights = {
                    name: numpy_to_torch(tensor)
                    for name, tensor in converted_weights.items()
                }

                # Bind the converted weights against the TRTLLM model
                model.load(converted_weights)

                # Write ranked-checkpoints
                to_safetensors(
                    converted_weights,
                    engines_folder / f"rank{model_config.mapping.rank}.safetensors",
                )

            # Write global config
            model_config.save_pretrained(engines_folder)

        # We are freeing memory used by the HF Model to let the engine build goes forward
        del hf_model
        torch.cuda.empty_cache()

        # Build
        engine_builder = LocalEngineBuilder(model_config, engines_folder)
        engine_builder.build(engine_config)

        return [engines_folder], [engines_folder.relative_to(local_path)]

    @classmethod
    def _from_pretrained(
        cls: Type[T],
        *,
        model_id: str,
        config: Dict[str, Any],
        revision: Optional[str],
        cache_dir: Optional[Union[str, Path]],
        force_download: bool,
        proxies: Optional[Dict],
        resume_download: bool,
        local_files_only: bool,
        token: Optional[Union[str, bool]],
        **model_kwargs,
    ) -> T:
        if not isinstance(cls, SupportsTensorrtConversion):
            raise ValueError(
                f"{cls} doesn't support converting from Hugging Face Hub model."
                " Please open up an issue at https://github.com/huggingface/optimum-nvidia/issues"
            )

        # Check if we are using a local path
        if not (local_path := Path(model_id)).exists():
            LOGGER.debug(
                f"Loading potential prebuilt engines from the Hub ({model_id}@{revision})"
            )

            # Let's retrieve the weights for this model
            # NOTE: We use `snapshot_download` to be able to provide a custom user-agent
            # NOTE: maybe we can do the same with `from_pretrained`
            local_path = HuggingFaceHubModel.retrieve_snapshot_from_hub(
                model_id,
                revision,
                cache_dir,
                force_download,
                proxies,
                resume_download,
                local_files_only,
                token,
                prebuilt_engines_only=True,
            )

        # Look for prebuilt engine files, if none found, we convert and build
        engines_folders, relative_paths_engines_folders = find_prebuilt_engines(
            local_path
        )
        if len(engines_folders) == 0:
            LOGGER.info(
                f"No engine file found in {local_path}, converting and building engines"
            )

            # If local_path exists and is not empty we have a local snapshot
            if (
                not local_path.exists()
                or len(list(local_path.glob("*.safetensors"))) == 0
            ):
                LOGGER.debug(
                    f"Loading original transformers weights from the Hub ({model_id}@{revision})"
                )

                local_path = HuggingFaceHubModel.retrieve_snapshot_from_hub(
                    model_id,
                    revision,
                    cache_dir,
                    force_download,
                    proxies,
                    resume_download,
                    local_files_only,
                    token,
                    prebuilt_engines_only=False,
                )

            engines_folders, relative_paths_engines_folders = cls.convert_and_build(
                local_path, config, **model_kwargs
            )
        else:
            LOGGER.info(f"Found pre-built engines at: {engines_folders}")

        try:
            generation_config = GenerationConfig.from_pretrained(
                model_id,
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                local_files_only=local_files_only,
                token=token,
            )
        except OSError:
            generation_config = None

        transformers_config = AutoConfig.from_pretrained(
            model_id,
            revision=revision,
            cache_dir=cache_dir,
            force_download=force_download,
            proxies=proxies,
            resume_download=resume_download,
            local_files_only=local_files_only,
            token=token,
        )

        model = cls(
            engines_folders,
            gpus_per_node=model_kwargs.pop("gpus_per_node", 1),
            use_cuda_graph=model_kwargs.pop("use_cuda_graph", False),
            generation_config=generation_config,
            transformers_config=transformers_config,
        )

        setattr(model, ATTR_TRTLLM_ENGINE_FOLDER, engines_folders)
        model._relative_path_engines_folders = relative_paths_engines_folders
        return model

    def _save_pretrained(self, save_directory: Path) -> None:
        if not hasattr(self, ATTR_TRTLLM_ENGINE_FOLDER):
            raise ValueError(
                "Unable to determine the root folder containing TensorRT-LLM engines. "
                "Please open-up an issue at https://github.com/huggingface/optimum-nvidia"
            )

        self.transformers_config.save_pretrained(save_directory)
        if self.generation_config is not None:
            self.generation_config.save_pretrained(save_directory)

        # Retrieve the folder
        engines_folders = getattr(self, ATTR_TRTLLM_ENGINE_FOLDER)
        for i in range(len(engines_folders)):
            engine_folder = Path(engines_folders[i])
            save_subfolder = self._relative_path_engines_folders[i]

            LOGGER.debug(f"Saving engines at {save_directory}")

            save_engine_directory = save_directory / save_subfolder
            save_engine_directory.mkdir(exist_ok=True, parents=True)

            # Move the engine(s)
            for engine in iglob(FILE_TRTLLM_ENGINE_PATTERN, root_dir=engine_folder):
                LOGGER.debug(
                    f"Moving file {engine_folder / engine} to {save_engine_directory / engine}"
                )
                shutil.copyfile(engine_folder / engine, save_engine_directory / engine)

            # Move the configuration
            if (config_path := engine_folder / "config.json").exists():
                shutil.copyfile(config_path, save_engine_directory / "config.json")
            else:
                LOGGER.warning(
                    f"No config.json found at {config_path}. It might not be possible to reload the engines."
                )

            if (original_config_path := engine_folder / ".." / "config.json").exists():
                shutil.copyfile(original_config_path, save_directory / "config.json")

    @staticmethod
    def convert_config_to_trtllm(
        cls: Type[SupportsTensorrtConversion],
        config: Union[PretrainedConfig, Dict[str, Any]],
        config_class: Optional[TensorRTConfig] = None,
        **additional_params,
    ) -> TensorRTConfig:
        """
        Convert a configuration initially generated by various Hugging Face libraries like transformers, diffusers, etc.
        to TensorRT-LLM `tensorrt_llm.modeling_utils.PretrainedConfig`.
        :param cls: The target class to allocate the model
        :param config: The original library configuration file
        :param config_class: An optional custom TensorRTConfig subclass to load into.
        :param additional_params:
        :return: `tensorrt_llm.modeling_utils.PretrainedConfig`
        """
        if config_class is None:
            config_class = cls.MODEL_CONFIG

        trt_config = config_class.from_config(config)
        if hasattr(trt_config, "check_config"):
            trt_config.check_config()

        return trt_config

    @staticmethod
    def retrieve_snapshot_from_hub(
        model_id: str,
        revision: Optional[str],
        cache_dir: Optional[Union[str, Path]],
        force_download: bool,
        proxies: Optional[Dict],
        resume_download: bool,
        local_files_only: bool,
        token: Optional[Union[str, bool]],
        prebuilt_engines_only: bool = False,
    ) -> Path:
        """

        :param model_id:
        :param revision:
        :param cache_dir:
        :param force_download:
        :param proxies:
        :param resume_download:
        :param local_files_only:
        :param token:
        :param prebuilt_engines_only:
        :return:
        """
        patterns = (
            HUB_TRTLLM_ENGINE_PATTERNS
            if prebuilt_engines_only
            else HUB_SAFETENSORS_PATTERNS
        )
        patterns += ["config.json"]
        local_path = snapshot_download(
            repo_id=model_id,
            revision=revision,
            cache_dir=cache_dir,
            allow_patterns=patterns,
            force_download=force_download,
            proxies=proxies,
            resume_download=resume_download,
            local_files_only=local_files_only,
            token=token,
            user_agent=get_user_agent(),
        )

        return Path(local_path)



import random
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Optional, Union

import numpy as np
import torch
from ammo.torch import quantization as atq
from datasets import Dataset
from transformers import PreTrainedTokenizer
from transformers.utils.quantization_config import QuantizationConfigMixin

from optimum.nvidia.lang import DataType
from optimum.nvidia.quantization.datasets import get_dataset


dtype = Union[str, torch.dtype]
TORCH_FLOAT8 = {torch.float8_e4m3fn, torch.float8_e5m2}


KV_CACHE_CFG = {
    "*.query_key_value.output_quantizer": {"num_bits": 8, "axis": None, "enable": True},
    "*.Wqkv.output_quantizer": {"num_bits": 8, "axis": None, "enable": True},
    "*.W_pack.output_quantizer": {"num_bits": 8, "axis": None, "enable": True},
    "*.c_attn.output_quantizer": {"num_bits": 8, "axis": None, "enable": True},
    "*.k_proj.output_quantizer": {"num_bits": 8, "axis": None, "enable": True},
    "*.v_proj.output_quantizer": {"num_bits": 8, "axis": None, "enable": True},
}


class QuantizationMethod(str, Enum):
    FLOAT8 = "fp8"


class AmmoQuantizationConfig(ABC, QuantizationConfigMixin):
    def __init__(
        self,
        quant_method: QuantizationMethod,
        with_quantized_kv_cache: bool = False,
        with_quantized_lm_head: bool = False,
        calibration_data: Optional[Dataset] = None,
    ):
        super().__init__(quant_method)

        self._with_quantized_kv_cache: bool = with_quantized_kv_cache
        self._with_quantized_lm_head: bool = with_quantized_lm_head
        self._calibration_dataset = calibration_data

    @property
    @abstractmethod
    def weight_dtype(self) -> torch.dtype:
        raise NotImplementedError("AmmoQuantizationConfig::weight_dtype is abstract.")

    @property
    @abstractmethod
    def has_quantized_kv_cache(self) -> bool:
        raise NotImplementedError(
            "AmmoQuantizationConfig::has_quantized_kv_cache is abstract."
        )

    @property
    def has_calibration_dataset(self) -> bool:
        return self._calibration_dataset is not None

    @property
    @abstractmethod
    def requires_calibration(self) -> bool:
        raise NotImplementedError(
            "AmmoQuantizationConfig::requires_calibration is abstract."
        )

    @property
    def calibration_dataset(self) -> Optional[Dataset]:
        return self._calibration_dataset

    @abstractmethod
    def as_ammo_config(self) -> Dict[str, Any]:
        raise NotImplementedError("AmmoQuantizationConfig::as_ammo_config is abstract.")

    def to_dict(self) -> Dict[str, Any]:
        return {
            "quant_method": self.quant_method,
            "with_kv_cache": self._with_quantized_kv_cache,
            "with_lm_head": self._with_quantized_lm_head,
        }

    def to_diff_dict(self) -> Dict[str, Any]:
        return self.to_dict()


class Float8QuantizationConfig(AmmoQuantizationConfig):
    def __init__(
        self,
        with_quantized_kv_cache: bool = False,
        with_quantized_lm_head: bool = False,
        calibration_data: Optional[Dataset] = None,
    ):
        super().__init__(
            QuantizationMethod.FLOAT8,
            with_quantized_kv_cache,
            with_quantized_lm_head,
            calibration_data,
        )

    @property
    def weight_dtype(self) -> torch.dtype:
        return torch.float8_e4m3fn

    @property
    def has_quantized_kv_cache(self) -> bool:
        return self._with_quantized_kv_cache

    @property
    def has_quantized_lm_head(self) -> bool:
        return self._with_quantized_lm_head

    @property
    def requires_calibration(self) -> bool:
        return True

    def as_ammo_config(self) -> Dict[str, Any]:
        cfg = atq.FP8_DEFAULT_CFG.copy()
        quant_config = cfg["quant_cfg"]

        if self.has_quantized_kv_cache:
            quant_kv_cache_config = KV_CACHE_CFG.copy()
            for value in quant_kv_cache_config.values():
                value.update(num_bits=(4, 3))

            quant_config.update(quant_kv_cache_config)

        quant_config["*lm_head*"] = {"enable": self.has_quantized_lm_head}

        return cfg


class AutoQuantizationConfig:
    @classmethod
    def from_dict(cls, kwargs):
        return cls.from_description(**kwargs)

    @classmethod
    def from_description(
        cls,
        weight: dtype,
        activation: dtype,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        dataset: Optional[Union[str, Dataset]] = None,
        split: Optional[str] = "train",
        num_samples: int = 512,
        max_sequence_length: int = 1024,
        seed: int = 2016,
        device: Union[str, torch.device] = "cpu",
    ):
        random.seed(seed)
        np.random.seed(seed)
        torch.random.manual_seed(seed)

        if isinstance(weight, str):
            weight = DataType(weight).to_torch()

        if isinstance(activation, str):
            activation = DataType(activation).to_torch()

        if isinstance(dataset, str):
            dataset = get_dataset(
                dataset,
                tokenizer,
                num_samples,
                seqlen=max_sequence_length,
                split=split,
                seed=seed,
                device=device,
            )
        else:
            raise ValueError("Providing custom dataset is not yet supported")

        # float8 case
        if weight in TORCH_FLOAT8:
            return Float8QuantizationConfig(
                with_quantized_kv_cache=activation in TORCH_FLOAT8,
                with_quantized_lm_head=False,
                calibration_data=dataset,
            )
        else:
            raise NotImplementedError(
                f"Quantization(weight= {weight}, activation={activation}) is not supported yet."
            )



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from itertools import chain
from logging import getLogger
from pathlib import Path
from subprocess import PIPE, STDOUT, run
from typing import Any, Dict

from optimum.nvidia import TensorRTConfig
from optimum.nvidia.builder.config import EngineConfig
from optimum.nvidia.utils.patching import BuilderPatcher


LOGGER = getLogger()
CLI_PLUGIN_NAMES = {
    # Plugins
    "bert_attention_plugin",
    "gpt_attention_plugin",
    "gemm_plugin",
    "lookup_plugin",
    "lora_plugin",
    "moe_plugin",
    # Features
    "context_fmha",
    "context_fmha_fp32_acc",
    "paged_kv_cache",
    "remove_input_padding",
    "use_custom_all_reduce",
    "multi_block_mode",
    "enable_xqa",
    "attention_qk_half_accumulation",
    "tokens_per_block",
    "use_paged_context_fmha",
    "use_context_fmha_for_generation",
}


def process_plugin_flag(_: str, value: Any) -> str:
    if isinstance(value, bool):
        return "enable" if value else "disable"
    else:
        return value


class LocalEngineBuilder:
    TRTLLM_BUILD_EXEC = "trtllm-build"

    @staticmethod
    def build_cli_command(
        root: Path, model_config: TensorRTConfig, build_config: EngineConfig
    ) -> Dict[str, Any]:
        workload_params = {
            "--max_batch_size": build_config.workload_profile.max_batch_size,
            "--max_input_len": build_config.workload_profile.max_input_len,
            "--max_output_len": build_config.workload_profile.max_output_len,
        }

        generation_params = {
            "--max_beam_width": build_config.generation_profile.num_beams,
        }

        if build_config.generation_profile.max_draft_length >= 1:
            generation_params["--max_draft_len"] = (
                build_config.generation_profile.max_draft_length
            )

        plugins_params = {
            f"--{name}": process_plugin_flag(name, value)
            for name in CLI_PLUGIN_NAMES
            if (value := getattr(build_config.plugins_config, name)) is not None
        }

        build_params = {
            "--checkpoint_dir": root,
            "--output_dir": root,
            "--model_config": root / "config.json",
            "--builder_opt": build_config.optimisation_level,
            "--logits_dtype": build_config.logits_dtype,
        }

        if hasattr(model_config, "trt_model_class") and hasattr(
            model_config, "trt_model_file"
        ):
            build_params["--model_cls_file"] = model_config.trt_model_file
            build_params["--model_cls_name"] = model_config.trt_model_class

        if model_config.supports_strong_typing():
            build_params["--strongly_typed"] = None

        return build_params | generation_params | workload_params | plugins_params

    def __init__(self, config: TensorRTConfig, output_folder: Path):
        self._config = config
        self._output_folder = output_folder

    def build(self, config: EngineConfig):
        cli_params = LocalEngineBuilder.build_cli_command(
            self._output_folder, self._config, config
        )
        cli_params_list = [str(t) for t in chain.from_iterable(cli_params.items())]
        cli_params_list = [i for i in cli_params_list if i != "None"]

        LOGGER.info(f"trtllm-build parameters: {cli_params_list}")

        for rank in range(self._config.mapping.world_size):
            ranked_checkpoint = f"rank{rank}.safetensors"
            if not (self._output_folder / ranked_checkpoint).exists():
                raise ValueError(
                    f"Missing rank-{rank} checkpoints (rank{rank}.safetensors), cannot build."
                )

        # TODO: Remove BuilderPatcher once TensorRT-LLM updates its codebase to allow to disable `optimize(network)`.
        with BuilderPatcher():
            # Run the build
            result = run(
                [LocalEngineBuilder.TRTLLM_BUILD_EXEC] + cli_params_list,
                stdout=PIPE,
                stderr=STDOUT,
            )

        if result.returncode != 0:
            LOGGER.warning(
                f"trtllm-build stdout: {result.stdout.decode('utf-8') if result.stdout is not None else None}"
            )
            LOGGER.warning(
                f"trtllm-build stderr: {result.stderr.decode('utf-8') if result.stderr is not None else None}"
            )

            raise ValueError(
                f"Compilation failed ({result.returncode}), "
                "please open up an issue at https://github.com/huggingface/optimum-nvidia"
            )  # TODO: change with proper error



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

DEFAULT_HF_HUB_TRT_REVISION: str = "trt-llm"

DEFAULT_ENGINE_FOLDER = ".engine"
TENSORRT_ENGINE_EXT = "engine"
TENSORRT_TIMINGS_FILE = "timings.cache"
OPTIMUM_NVIDIA_CONFIG_FILENAME = "build"
OPTIMUM_NVIDIA_CONFIG_FILE = f"{OPTIMUM_NVIDIA_CONFIG_FILENAME}.json"

# Nvidia Compute Capabilities
SM_VOLTA = 70
SM_TURING = 75
SM_AMPERE = 80
SM_AMPERE_NEXT = 86
SM_ADA_LOVELACE = 89
SM_HOPPER = 90



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from logging import getLogger

from accelerate import cpu_offload_with_hook
from accelerate.hooks import remove_hook_from_module
from torch.nn import Module


LOGGER = getLogger(__name__)


# Copied from optimum.utils._modeling_utils
def recurse_getattr(obj, attr: str):
    """
    Recursive `getattr`.

    Args:
        obj:
            A class instance holding the attribute.
        attr (`str`):
            The attribute that is to be retrieved, e.g. 'attribute1.attribute2'.
    """

    def _getattr(obj, attr):
        return getattr(obj, attr)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def maybe_offload_weights_to_cpu(model: Module):
    if hasattr(model, "hf_device_map"):
        devices = set(model.hf_device_map.values())
        if "disk" in devices:
            raise ValueError("disk offload is not supported with quantization")
        if "cpu" in devices and len(model.hf_device_map) > 1:
            hook = None
            for name, device in model.hf_device_map.items():
                if device == "cpu":
                    LOGGER.debug(f"Offloading {name} to device {device}")
                    module = recurse_getattr(model, name)
                    remove_hook_from_module(module, recurse=True)
                    module, hook = cpu_offload_with_hook(module, prev_module_hook=hook)

    return model



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import functools

from .constants import (
    DEFAULT_ENGINE_FOLDER,
    DEFAULT_HF_HUB_TRT_REVISION,
    OPTIMUM_NVIDIA_CONFIG_FILE,
    OPTIMUM_NVIDIA_CONFIG_FILENAME,
    TENSORRT_ENGINE_EXT,
    TENSORRT_TIMINGS_FILE,
)
from .env import parse_flag_from_env
from .hub import get_user_agent
from .nvml import has_float8_support
from .offload import maybe_offload_weights_to_cpu
from .onnx import to_onnx


def rgetattr(obj, attr):
    """
    Recursively get object attribute
    :param obj: The root object we want to retrieve nested attribute
    :param attr: The attribute path where nested attribute are comma delimited
    :return:
    """

    def _getattr(obj, attr):
        return getattr(obj, attr)

    return functools.reduce(_getattr, [obj] + attr.split("."))



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import pytest
import tensorrt as trt
import torch

from optimum.nvidia.lang import DataType


@pytest.mark.parametrize(
    "literal_dtype,dtype",
    [
        ("bool", torch.bool),
        ("uint8", torch.uint8),
        ("int8", torch.int8),
        # ("int16", torch.int16),
        ("int32", torch.int32),
        ("int64", torch.int64),
        ("float8", torch.float8_e4m3fn),  # Change this when supported
        ("float16", torch.float16),
        ("bfloat16", torch.bfloat16),
        ("float32", torch.float32),
    ],
)
def test_convert_str_to_torch(literal_dtype: str, dtype):
    assert DataType(literal_dtype).to_torch() == dtype


@pytest.mark.parametrize(
    "literal_dtype,dtype",
    [
        ("uint8", trt.uint8),
        ("int8", trt.int8),
        ("int32", trt.int32),
        ("int64", trt.int64),
        ("float8", trt.fp8),
        ("float16", trt.float16),
        ("bfloat16", trt.bfloat16),
        ("float32", trt.float32),
    ],
)
def test_convert_str_to_tensorrt(literal_dtype: str, dtype):
    assert DataType(literal_dtype).to_trt() == dtype



import inspect

from tensorrt_llm import Builder


class BuilderPatcher:
    # TODO: Remove this once available natively in TensorRT-LLM - `optimize(network)` lowers predictive performance for Whisper.

    def __init__(self):
        self.builder_path = inspect.getfile(Builder)

    def __enter__(self):
        with open(self.builder_path, "r") as file:
            lines = file.readlines()

        for i in range(len(lines)):
            if (
                i > 0
                and lines[i] == "    optimize(network)\n"
                and "# PATCHED" not in lines[i - 1]
            ):
                lines[i] = (
                    "    if getattr(model.config, 'optimize_network', True):\n        # PATCHED\n        optimize(network)\n    else:\n        logger.info('Network optimization disabled during build.')\n"
                )

        with open(self.builder_path, "w") as f:
            f.writelines(lines)

    def __exit__(self, exc_type, exc_value, traceback):
        with open(self.builder_path, "r") as file:
            lines = file.readlines()

        for i in range(len(lines)):
            if lines[i] == "        optimize(network)\n":
                lines[i - 1] = ""
                lines[i - 2] = ""
                lines[i] = "    optimize(network)\n"
                lines[i + 1] = ""
                lines[i + 2] = ""

        with open(self.builder_path, "w") as f:
            f.writelines(lines)



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from logging import getLogger
from typing import Optional, Union

import torch
from tensorrt_llm.models import PretrainedConfig
from tensorrt_llm.plugin import PluginConfig

from optimum.nvidia.lang import DataType


LOGGER = getLogger()
SUPPORTED_LOGITS_DTYPE = {"float32", "float16"}


@dataclass
class InferenceProfile:
    max_batch_size: int
    max_input_len: int
    max_output_len: int


@dataclass
class GenerationProfile:
    num_beams: int
    max_draft_length: int


@dataclass
class ShardingProfile:
    tensor_parallelism: int = 1
    pipeline_parallelism: int = 1
    world_size: int = 1
    gpus_per_node: int = 1


@dataclass
class EngineConfig:
    """
    Represent all the parameters required to tune and build the final TRTLLM engine(s)
    """

    optimisation_level: int
    strongly_typed: bool
    logits_dtype: str
    workload_profile: InferenceProfile
    generation_profile: GenerationProfile
    sharding_profile: ShardingProfile
    plugins_config: PluginConfig


class EngineConfigBuilder:
    @staticmethod
    def from_dict(config: PretrainedConfig, **additional_params):
        builder = EngineConfigBuilder(config)

        # Define the data type to export the logits
        builder.logits_as(additional_params.pop("logits_dtype", config.logits_dtype))

        # Workload related
        max_batch_size = additional_params.pop("max_batch_size", 1)
        max_prompt_length = additional_params.pop("max_prompt_length", 128)
        max_new_tokens = (
            additional_params.pop("max_output_length", config.max_position_embeddings)
            - max_prompt_length
        )

        if max_new_tokens < 1:
            raise ValueError(
                "Unable to build the engine because the generation would lead to max_num_tokens < 1. ("
                f"max_prompt_length = {max_prompt_length}, "
                f"max_position_embeddings={config.max_position_embeddings}, "
                f"max_new_tokens={max_new_tokens}"
                ")"
            )

        builder.with_inference_profile(
            max_batch_size, max_prompt_length, max_new_tokens
        )

        # Generation related
        builder.with_generation_profile(additional_params.pop("num_beams", 1))

        # Speculative decoding
        if "max_speculated_draft_length" in additional_params:
            builder.with_speculated_decoding(
                additional_params.pop("max_speculated_draft_length")
            )

        return builder

    def __init__(self, config: PretrainedConfig):
        self._config = config

        self._optimisation_level: int = 3
        self._logits_dtype = config.logits_dtype
        self._strongly_typed: bool = False
        self._sharding_profile: ShardingProfile = ShardingProfile()
        self._workload_profile: Optional[InferenceProfile] = None
        self._generation_profile: Optional[GenerationProfile] = None
        self._plugin_config: Optional[PluginConfig] = None

    def strongly_typed(self) -> "EngineConfigBuilder":
        self._strongly_typed = True
        LOGGER.info("Defined engine as strongly typed")
        return self

    def shard(
        self,
        tensor_parallelism: int = 1,
        pipeline_parallelism: int = 1,
        world_size: int = 1,
        gpus_per_node: int = 1,
    ) -> "EngineConfigBuilder":
        self._sharding_profile = ShardingProfile(
            tensor_parallelism, pipeline_parallelism, world_size, gpus_per_node
        )
        LOGGER.debug(f"Defined sharding profile as: {self._sharding_profile}")

        return self

    def with_optimisation_level(self, level: int) -> "EngineConfigBuilder":
        if level < 1:
            raise ValueError(f"level should be >= 1 (got: {level})")
        self._optimisation_level = level
        LOGGER.info(f"Defined optimisation level to {self._optimisation_level}")
        return self

    def logits_as(
        self, dtype: Union[str, torch.dtype, DataType]
    ) -> "EngineConfigBuilder":
        if isinstance(dtype, torch.dtype):
            dtype = DataType.from_torch(dtype)

        if isinstance(dtype, DataType):
            dtype = dtype.value

        if dtype not in SUPPORTED_LOGITS_DTYPE:
            dtype = "float32"

        self._logits_dtype = dtype
        LOGGER.info(f"Defined logits dtype to: {self._logits_dtype}")
        return self

    def with_inference_profile(
        self, max_batch_size: int, max_prompt_length: int, max_new_tokens: int
    ) -> "EngineConfigBuilder":
        if max_batch_size < 1:
            raise ValueError(f"max_batch_size should be >= 1 (got: {max_batch_size})")

        if max_prompt_length < 1:
            raise ValueError(
                f"max_prompt_length should be >= 1 (got: {max_batch_size})"
            )

        if max_prompt_length >= self._config.max_position_embeddings:
            raise ValueError(
                f"max_prompt_length should be shorter than the maximum length supported by the model."
                f" (got: {max_prompt_length} and"
                f" maximum sequence length supported by the model is {self._config.max_position_embeddings})"
            )

        if max_new_tokens < 1:
            raise ValueError(f"max_new_tokens should be >= 1 (got: {max_new_tokens})")

        if max_new_tokens > self._config.max_position_embeddings:
            raise ValueError(
                f"max_new_tokens should be shorter than the maximum length supported by the model."
                f" (got: {max_new_tokens} and"
                f" maximum sequence length supported by the model is {self._config.max_position_embeddings})"
            )

        self._workload_profile = InferenceProfile(
            max_batch_size, max_prompt_length, max_new_tokens
        )
        LOGGER.info(f"Defined engine inference profile: {self._workload_profile}")
        return self

    def with_generation_profile(self, num_beams: int) -> "EngineConfigBuilder":
        if num_beams < 1:
            raise ValueError(f"num_beams should be >= 1 (got: {num_beams})")

        self._generation_profile = GenerationProfile(num_beams, -1)
        LOGGER.info(f"Defined engine generation profile: {self._generation_profile}")
        return self

    def with_speculated_decoding(self, max_draft_length: int) -> "EngineConfigBuilder":
        if max_draft_length < 1:
            raise ValueError(
                f"max_draft_length should be >= 1 (got: {max_draft_length})"
            )

        if self._generation_profile is None:
            raise ValueError(
                "You should specify generation profile first. "
                "Please use EngineConfigBuilder.with_generation_profile()"
            )

        self._generation_profile = GenerationProfile(
            self._generation_profile.num_beams, max_draft_length
        )
        LOGGER.info(
            f"Defined engine generation profile with speculation: {self._generation_profile}"
        )
        return self

    def with_plugins_config(self, plugin_config: PluginConfig) -> "EngineConfigBuilder":
        self._plugin_config = plugin_config
        LOGGER.info(f"Defined plugins config: {plugin_config}")
        return self

    def validate(self) -> bool:
        if self._workload_profile is None:
            raise ValueError(
                "You need to set an inference profile. Use EngineConfigBuilder.with_inference_profile()."
            )

        if self._generation_profile is None:
            raise ValueError(
                "You need to set a generation profile. Use EngineConfigBuilder.with_generation_profile()."
            )

        if self._plugin_config is None:
            raise ValueError(
                "You need to set a plugin profile. Use EngineConfigBuilder.with_plugins_config()."
            )

        max_generated_length = (
            self._workload_profile.max_input_len
            + self._workload_profile.max_output_len
            - 1
        )
        if max_generated_length > self._config.max_position_embeddings:
            raise ValueError(
                "max_prompt_length + max_new_tokens should be lesser or equals "
                "to the maximum length supported by the model (got "
                f"max_prompt_length={self._workload_profile.max_input_len}, "
                f"max_new_tokens={self._workload_profile.max_output_len},"
                f"{self._workload_profile.max_input_len + self._workload_profile.max_output_len}"
                f" > {self._config.max_position_embeddings}"
                ")"
            )

        return True

    def build(self) -> EngineConfig:
        self.validate()
        return EngineConfig(
            optimisation_level=self._optimisation_level,
            sharding_profile=self._sharding_profile,
            strongly_typed=self._strongly_typed,
            logits_dtype=self._logits_dtype,
            workload_profile=self._workload_profile,
            generation_profile=self._generation_profile,
            plugins_config=self._plugin_config,
        )



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from sys import version as pyversion

from pynvml import nvmlDeviceGetCount, nvmlInit, nvmlSystemGetDriverVersion

from optimum.nvidia.utils import parse_flag_from_env

from ..version import __version__
from .nvml import get_device_compute_capabilities


USER_AGENT_BASE = [f"optimum/nvidia/{__version__}", f"python/{pyversion.split()[0]}"]


@functools.cache
def get_user_agent() -> str:
    """
    Get the library user-agent when calling the hub
    :return:
    """
    ua = USER_AGENT_BASE.copy()

    # Nvidia driver / devices
    try:
        nvmlInit()
        ua.append(f"nvidia/{nvmlSystemGetDriverVersion()}")

        num_gpus = nvmlDeviceGetCount()
        if num_gpus > 0:
            sm = []
            for device_idx in range(num_gpus):
                compute_capabilities = get_device_compute_capabilities(device_idx)
                if compute_capabilities:
                    major, minor = compute_capabilities
                    sm.append(f"{major}{minor}")

            ua.append(f"gpus/{num_gpus}")
            ua.append(f"sm/{'|'.join(sm)}")
    except (RuntimeError, ImportError):
        ua.append("nvidia/unknown")

    # Torch / CUDA related version, (from torch)
    try:
        from torch import __version__ as pt_version
        from torch.version import cuda, cudnn

        ua.append(f"cuda/{cuda}")
        ua.append(f"cudnn/{cudnn}")
        ua.append(f"torch/{pt_version}")
    except ImportError:
        pass

    # transformers version
    try:
        from transformers import __version__ as tfrs_version

        ua.append(f"transformers/{tfrs_version}")
    except ImportError:
        pass

    # TRTLLM version
    try:
        from tensorrt_llm._utils import trt_version

        ua.append(f"tensorrt/{trt_version()}")
    except ImportError:
        pass

    # Add a flag for CI
    if parse_flag_from_env("OPTIMUM_NVIDIA_IS_CI", False):
        ua.append("is_ci/true")
    else:
        ua.append("is_ci/false")

    return "; ".join(ua)



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from enum import Enum
from typing import List

import torch


class DataType(Enum):
    FLOAT32 = "float32"
    FLOAT16 = "float16"
    BFLOAT16 = "bfloat16"
    FLOAT8 = "float8"
    INT64 = "int64"
    INT32 = "int32"
    INT8 = "int8"
    UINT8 = "uint8"
    BOOL = "bool"

    @staticmethod
    def from_torch(dtype: torch.dtype) -> "DataType":
        if dtype == torch.float32:
            return DataType.FLOAT32
        elif dtype == torch.float16:
            return DataType.FLOAT16
        elif dtype == torch.bfloat16:
            return DataType.BFLOAT16
        elif dtype == torch.float8_e4m3fn:
            return DataType.FLOAT8
        elif dtype == torch.int64:
            return DataType.INT64
        elif dtype == torch.int32:
            return DataType.INT32
        elif dtype == torch.int8:
            return DataType.INT8
        elif dtype == torch.uint8:
            return DataType.UINT8
        elif dtype == torch.bool:
            return DataType.BOOL
        else:
            raise ValueError(f"Unknown torch.dtype {dtype}")

    def to_trt(self) -> "DataType":
        """
        Convert textual dtype representation to their TensorRT counterpart
        :return: Converted dtype if equivalent is found
        :raise ValueError if provided dtype doesn't have counterpart
        """
        import tensorrt as trt

        if self == DataType.FLOAT32:
            return trt.DataType.FLOAT
        elif self == DataType.FLOAT16:
            return trt.DataType.HALF
        elif self == DataType.BFLOAT16:
            return trt.DataType.BF16
        elif self == DataType.FLOAT8:
            return trt.DataType.FP8
        elif self == DataType.INT8:
            return trt.DataType.INT8
        elif self == DataType.UINT8:
            return trt.DataType.UINT8
        elif self == DataType.INT32:
            return trt.DataType.INT32
        elif self == DataType.INT64:
            return trt.DataType.INT64
        elif self == DataType.BOOL:
            return trt.DataType.BOOL
        else:
            raise ValueError(f"Unknown value {self}")

    def to_torch(self):
        """
        Convert textual dtype representation to their Torch counterpart
        :return: Converted dtype if equivalent is found
        :raise ValueError if provided dtype doesn't have counterpart
        """
        import torch

        if self == DataType.FLOAT32:
            return torch.float32
        elif self == DataType.FLOAT16:
            return torch.float16
        elif self == DataType.BFLOAT16:
            return torch.bfloat16
        elif self == DataType.FLOAT8:
            return torch.float8_e4m3fn
        elif self == DataType.INT8:
            return torch.int8
        elif self == DataType.UINT8:
            return torch.uint8
        elif self == DataType.INT32:
            return torch.int32
        elif self == DataType.INT64:
            return torch.int64
        elif self == DataType.BOOL:
            return torch.bool
        else:
            raise ValueError(f"Unknown value {self}")

    @staticmethod
    def values() -> List[str]:
        return [item.value for item in DataType]



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from enum import Enum
from typing import Dict, List, Union

import torch
from transformers import PreTrainedTokenizer, TensorType

from optimum.nvidia import AutoModelForCausalLM
from optimum.nvidia.runtime import CausalLM

from .base import Pipeline


class ReturnType(Enum):
    TENSORS = 0
    NEW_TEXT = 1
    FULL_TEXT = 2


class TextGenerationPipeline(Pipeline):
    TARGET_FACTORY = AutoModelForCausalLM

    __slots__ = (
        "tokenizer",
        "_runtime",
        "_bos_token_id",
        "_eos_token_id",
        "_pad_token_id",
    )

    def __init__(self, model: CausalLM, tokenizer: PreTrainedTokenizer):
        super().__init__()

        if tokenizer.eos_token and not tokenizer.pad_token:
            tokenizer.pad_token = tokenizer.eos_token

        self.tokenizer = tokenizer
        self._runtime = model

        self._bos_token_id = tokenizer.bos_token_id
        self._eos_token_id = tokenizer.eos_token_id
        self._pad_token_id = tokenizer.pad_token_id

    def __call__(self, inputs: Union[str, List[str]], **kwargs):
        (
            preprocess_params,
            forward_params,
            postprocess_params,
        ) = self._sanitize_parameters(**kwargs)
        model_inputs = self.preprocess(inputs, **preprocess_params)
        model_outputs = self._forward(model_inputs, **forward_params)
        outputs = self.postprocess(model_outputs, **postprocess_params)
        return outputs

    def _sanitize_parameters(
        self,
        return_full_text=None,
        return_tensors=None,
        return_text=None,
        return_type=None,
        clean_up_tokenization_spaces=None,
        prefix=None,
        handle_long_generation=None,
        stop_sequence=None,
        add_special_tokens=False,
        **generate_kwargs,
    ):
        preprocess_params = {"add_special_tokens": add_special_tokens}
        if prefix is not None:
            preprocess_params["prefix"] = prefix
        if prefix:
            prefix_inputs = self.tokenizer(
                prefix,
                padding=False,
                add_special_tokens=add_special_tokens,
                return_tensors=TensorType.PYTORCH,
            )
            generate_kwargs["prefix_length"] = prefix_inputs["input_ids"].shape[-1]

        if handle_long_generation is not None:
            if handle_long_generation not in {"hole"}:
                raise ValueError(
                    f"{handle_long_generation} is not a valid value for `handle_long_generation` parameter expected"
                    " [None, 'hole']"
                )
            preprocess_params["handle_long_generation"] = handle_long_generation

        preprocess_params.update(generate_kwargs)
        forward_params = generate_kwargs

        postprocess_params = {}
        if return_full_text is not None and return_type is None:
            if return_text is not None:
                raise ValueError(
                    "`return_text` is mutually exclusive with `return_full_text`"
                )
            if return_tensors is not None:
                raise ValueError(
                    "`return_full_text` is mutually exclusive with `return_tensors`"
                )
            return_type = (
                ReturnType.FULL_TEXT if return_full_text else ReturnType.NEW_TEXT
            )
        if return_tensors is not None and return_type is None:
            if return_text is not None:
                raise ValueError(
                    "`return_text` is mutually exclusive with `return_tensors`"
                )
            return_type = ReturnType.TENSORS
        if return_type is not None:
            postprocess_params["return_type"] = return_type
        if clean_up_tokenization_spaces is not None:
            postprocess_params["clean_up_tokenization_spaces"] = (
                clean_up_tokenization_spaces
            )

        if stop_sequence is not None:
            stop_sequence_ids = self.tokenizer.encode(
                stop_sequence, add_special_tokens=False
            )
            if len(stop_sequence_ids) > 1:
                warnings.warn(
                    "Stopping on a multiple token sequence is not yet supported on transformers. The first token of"
                    " the stop sequence will be used as the stop sequence string in the interim."
                )
            generate_kwargs["eos_token_id"] = stop_sequence_ids[0]

        return preprocess_params, forward_params, postprocess_params

    def _forward(self, model_inputs, **generate_kwargs):
        input_ids = model_inputs["input_ids"]
        prompt_text = model_inputs.pop("prompt_text")
        attention_mask = model_inputs.get("attention_mask", None)

        max_new_tokens = generate_kwargs.pop("max_new_tokens", -1)
        min_length = generate_kwargs.pop("min_length", -1)
        num_beams = generate_kwargs.pop("num_beams", 1)
        temperature = generate_kwargs.pop("temperature", 1.0)
        top_k = generate_kwargs.pop("top_k", 50)
        top_p = generate_kwargs.pop("top_p", 1.0)
        repetition_penalty = generate_kwargs.pop("repetition_penalty", 1.0)
        length_penalty = generate_kwargs.pop("length_penalty", 1.0)
        seed = generate_kwargs.pop("seed", 2017)

        # prefix_length = generate_kwargs.pop("prefix_length", 0)
        # If there is a prefix, we may need to adjust the generation length. Do so without permanently modifying
        # generate_kwargs, as some of the parameterization may come from the initialization of the pipeline.
        # if prefix_length > 0:
        #     has_max_new_tokens = "max_new_tokens" in generate_kwargs or (
        #             "generation_config" in generate_kwargs
        #             and generate_kwargs["generation_config"].max_new_tokens is not None
        #     )
        #     if not has_max_new_tokens:
        #         generate_kwargs["max_length"] = generate_kwargs.get("max_length") or self._runtime
        #         generate_kwargs["max_length"] += prefix_length
        #     has_min_new_tokens = "min_new_tokens" in generate_kwargs or (
        #             "generation_config" in generate_kwargs
        #             and generate_kwargs["generation_config"].min_new_tokens is not None
        #     )
        #     if not has_min_new_tokens and "min_length" in generate_kwargs:
        #         generate_kwargs["min_length"] += prefix_length

        # BS x BEAMS x SL
        generated_sequence, lengths = self._runtime.generate(
            input_ids=input_ids,
            attention_mask=attention_mask,
            max_new_tokens=max_new_tokens,
            min_length=min_length,
            num_beams=num_beams,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            repetition_penalty=repetition_penalty,
            length_penalty=length_penalty,
            seed=seed,
            bos_token_id=self._bos_token_id,
            eos_token_id=self._eos_token_id,
            pad_token_id=self._pad_token_id,
        )

        return {
            "generated_sequence": generated_sequence,
            "lengths": lengths,
            "input_ids": input_ids,
            "prompt_text": prompt_text,
        }

    def preprocess(
        self,
        prompt_text,
        prefix="",
        handle_long_generation=None,
        add_special_tokens=False,
        **generate_kwargs,
    ) -> Dict[str, torch.Tensor]:
        if isinstance(prompt_text, List):
            text = [prefix + prompt for prompt in prompt_text]
        else:
            text = prefix + prompt_text

        inputs = self.tokenizer(
            text,
            padding=False,
            add_special_tokens=add_special_tokens,
            return_tensors=TensorType.PYTORCH,
        )
        inputs["prompt_text"] = prompt_text

        return inputs

    def postprocess(
        self,
        model_outputs,
        return_type=ReturnType.FULL_TEXT,
        clean_up_tokenization_spaces=True,
    ):
        generated_sequence = model_outputs["generated_sequence"]
        # lengths = model_outputs["lengths"]
        # input_ids = model_outputs["input_ids"]
        # prompt_text = model_outputs["prompt_text"]
        generated_sequence = generated_sequence.cpu().numpy().tolist()
        records = []

        if return_type == ReturnType.TENSORS:
            return [
                {"generated_token_ids": generated for generated in generated_sequence}
            ]

        for sequence in generated_sequence:
            # Decode text
            beam_text = self.tokenizer.batch_decode(
                sequence,
                skip_special_tokens=True,
                clean_up_tokenization_spaces=clean_up_tokenization_spaces,
            )

            record = {"generated_text": beam_text}
            records.append(record)

        return records



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from logging import getLogger
from typing import NamedTuple, Optional, Tuple

from psutil import virtual_memory
from pynvml import (
    NVMLError,
    nvmlDeviceGetCudaComputeCapability,
    nvmlDeviceGetHandleByIndex,
    nvmlDeviceGetMemoryInfo,
    nvmlDeviceGetName,
    nvmlInit,
)


LOGGER = getLogger()


_NVML_INITIALIZED = False

SM_FP8_SUPPORTED = {89, 90}


MemoryInfo = NamedTuple(
    "MemoryInfo",
    [
        ("total", int),
        ("free", int),
        ("used", int),
    ],
)


def nvml_guard(f):
    @functools.wraps(f)
    def _ensure_init(*args, **kwargs):
        global _NVML_INITIALIZED
        if not _NVML_INITIALIZED:
            nvmlInit()
            _NVML_INITIALIZED = True

        return f(*args, **kwargs)

    return _ensure_init


@functools.cache
@nvml_guard
def get_device_compute_capabilities(device: int) -> Optional[Tuple[int, int]]:
    nvml_device_handle = nvmlDeviceGetHandleByIndex(device)
    return nvmlDeviceGetCudaComputeCapability(nvml_device_handle)


@functools.cache
@nvml_guard
def get_device_memory(device: int) -> Optional[int]:
    nvml_device_handle = nvmlDeviceGetHandleByIndex(device)
    mem_info = nvmlDeviceGetMemoryInfo(nvml_device_handle)
    return MemoryInfo(mem_info.total, mem_info.free, mem_info.used).total


@functools.cache
@nvml_guard
def get_device_count() -> int:
    import torch

    return torch.cuda.device_count()


@functools.cache
@nvml_guard
def get_device_name(device: int) -> Optional[Tuple[str, ...]]:
    try:
        nvml_device_handle = nvmlDeviceGetHandleByIndex(device)
        device_full_name = nvmlDeviceGetName(nvml_device_handle)
        return device_full_name.split()
    except NVMLError:
        return None


@functools.cache
@nvml_guard
def has_float8_support() -> bool:
    compute_capabilities = get_device_compute_capabilities(0)
    if compute_capabilities:
        compute_capabilities_ = compute_capabilities[0] * 10 + compute_capabilities[1]

        return compute_capabilities_ in SM_FP8_SUPPORTED
    else:
        LOGGER.warning(
            "Failed to retrieve the proper compute capabilities on the device"
        )
        return False


def get_max_memory():
    fraction_device_map = {
        device_id: get_device_memory(device_id) * 0.7
        for device_id in range(get_device_count())
    }

    cpu_device_map = {"cpu": virtual_memory().available * 0.8}

    return fraction_device_map | cpu_device_map



#  coding=utf-8
#  Copyright 2024 The HuggingFace Inc. team. All rights reserved.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import Union

import torch
from ammo.torch import export as ate
from ammo.torch import quantization as atq
from tensorrt_llm.quantization import QuantMode
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import PreTrainedModel
from transformers.quantizers import HfQuantizer
from transformers.utils.quantization_config import QuantizationConfigMixin

from optimum.nvidia.quantization.ammo import AmmoQuantizationConfig


LOGGER = getLogger(__name__)

_SUPPORTED_MODEL_ARCHITECTURES = {
    "llama": "llama",
    "mistral": "llama",
    "gemma": "gemma",
}


def get_quantization_algorithm_name(qconfig: QuantMode) -> str:
    if qconfig.has_fp8_qdq() or qconfig.has_fp8_kv_cache():
        return "fp8"
    elif qconfig.is_int4_weight_only():
        return "int4_awq"
    elif not qconfig.has_act_and_weight_quant():
        return "int8_sq"
    else:
        raise ValueError(f"Unable to determine quantization algorithm from: {qconfig}")


def infer_decoder_type(model: PreTrainedModel) -> str:
    if model.config.model_type in _SUPPORTED_MODEL_ARCHITECTURES:
        return _SUPPORTED_MODEL_ARCHITECTURES[model.config.model_type]

    else:
        raise ValueError(f"{model.config.model_type} is not supported yet")


class AmmoQuantizer(HfQuantizer):
    def __init__(
        self,
        quantization_config: QuantizationConfigMixin,
        artifact_path: Union[str, PathLike, Path],
        tensor_parallel_degree: int = 1,
        pipeline_parallel_degree: int = 1,
        export_tensorrt_llm_config: bool = True,
    ):
        if tensor_parallel_degree < 1:
            raise ValueError(
                f"tensor_parallel_degree should be >= 1 (got {tensor_parallel_degree})"
            )

        if pipeline_parallel_degree < 1:
            raise ValueError(
                f"pipeline_parallel_degree should be >= 1 (got {pipeline_parallel_degree})"
            )

        super().__init__(quantization_config=quantization_config)

        if not isinstance(artifact_path, Path):
            artifact_path = Path(artifact_path)

        self._artifact_path = artifact_path
        self._tp_degree = tensor_parallel_degree
        self._pp_degree = pipeline_parallel_degree
        self._export_tensorrt_llm_config = export_tensorrt_llm_config

    @property
    def is_serializable(self):
        return False

    @property
    def is_trainable(self):
        return False

    def _process_model_before_weight_loading(
        self, model, batch_size: int = 1, **kwargs
    ):
        assert isinstance(self.quantization_config, AmmoQuantizationConfig)
        qconfig = self.quantization_config

        if qconfig.requires_calibration:
            if not qconfig.has_calibration_dataset:
                raise ValueError("Float8 quantization requires a calibration dataset")

            with torch.inference_mode():

                def _loop():
                    data = DataLoader(
                        qconfig.calibration_dataset,
                        batch_size=batch_size,
                        pin_memory=True,
                        pin_memory_device="cuda:0",
                    )

                    for sample in tqdm(data):
                        inputs = {
                            name: tensor.to("cuda:0") for name, tensor in sample.items()
                        }
                        model(**inputs)

                atq.quantize(model, config=qconfig.as_ammo_config(), forward_loop=_loop)

    def _process_model_after_weight_loading(self, model, **kwargs):
        assert isinstance(self.quantization_config, AmmoQuantizationConfig)

        with torch.inference_mode():
            decoder_type = infer_decoder_type(model)
            ate.export_model_config(
                model=model,
                decoder_type=decoder_type,
                dtype=model.config.torch_dtype,
                export_dir=self._artifact_path,
                inference_tensor_parallel=self._tp_degree,
                inference_pipeline_parallel=self._pp_degree,
                export_tensorrt_llm_config=self._export_tensorrt_llm_config,
                export_npz=False,
            )



import torch
from transformers import (
    ForceTokensLogitsProcessor,
    SuppressTokensAtBeginLogitsProcessor,
    SuppressTokensLogitsProcessor,
)
from transformers.generation.logits_process import WhisperNoSpeechDetection


class TrtSuppressTokensLogitsProcessor(SuppressTokensLogitsProcessor):
    def __call__(self, step: int, input_ids: torch.Tensor, scores: torch.Tensor):
        scores = super().__call__(input_ids, scores)
        return scores


class TrtSuppressTokensAtBeginLogitsProcessor(SuppressTokensAtBeginLogitsProcessor):
    def __call__(self, step: int, input_ids: torch.Tensor, scores: torch.Tensor):
        scores = super().__call__(input_ids, scores)
        return scores


class TrtForceTokensLogitsProcessor(ForceTokensLogitsProcessor):
    def __call__(self, step: int, input_ids: torch.Tensor, scores: torch.Tensor):
        scores = super().__call__(input_ids, scores)
        return scores


class TrtWhisperNoSpeechDetection(WhisperNoSpeechDetection):
    def __call__(self, step: int, input_ids: torch.Tensor, scores: torch.Tensor):
        scores = super().__call__(input_ids, scores)
        return scores


LOGITS_PROCESSOR_MAP = {
    SuppressTokensLogitsProcessor: TrtSuppressTokensLogitsProcessor,
    SuppressTokensAtBeginLogitsProcessor: TrtSuppressTokensAtBeginLogitsProcessor,
    ForceTokensLogitsProcessor: TrtForceTokensLogitsProcessor,
    WhisperNoSpeechDetection: TrtWhisperNoSpeechDetection,
}
